'''
generic image and metric data class constructor
both use the same image for reference and transformed
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
from functools import lru_cache
from collections import namedtuple
import numpy as np
import pandas as pd
import imghdr
import IQM_Vis
from IQM_Vis.data_handlers import base_dataset_loader


# keep a track of all the cached functions so we can clear them easily
CACHED_FUNCTIONS = []
# custom decorator to cache functions and store which are cached
def cache_tracked(func):
    cached_func = lru_cache(maxsize=None)(func)
    CACHED_FUNCTIONS.append(cached_func)
    return cached_func


class cache_metric_call:
    ''' cache metric fucntions that have been calculated already
     to do this we need to convert numpy arrays to a hashable object (since arrs are mutable) 
     to achieve this we convert to a bytes array (input to __call__) then load this from buffer
     '''
    def __init__(self, metric):
        self.metric = metric
    
    @cache_tracked
    def __call__(self, ref, trans, **kwargs):
        # expect a hashable bytes array tuple as input with the data type and shape
        #  N.B. we need to copy the array since from buffer gives a read only array since
        #       it is a view of a bytes array (immutable)
        ref = np.frombuffer(ref.bytes, dtype=ref.dtype).reshape(ref.shape).copy()
        trans = np.frombuffer(trans.bytes, dtype=trans.dtype).reshape(trans.shape).copy()
        return self.metric(ref, trans, **kwargs)


class dataset_holder(base_dataset_loader):
    '''Stores images and metrics to communicate with the UI via the IQM-Vis data
       API.

       Args:
           image_list (list): list of image file paths
           metrics (dict): dictionary with keys of the metric names and values
                           of the callable metric function
           metric_images (dict): Optional dictionary with keys of the metric
                                 image names and values of the callable metric
                                 image function (Defaults to {})
           image_loader (function): Optional function which loads an image from
                                    a file path (Defaults to IQM_Vis.utils.load_image)
           image_post_processing (function): Optional function to apply after image
                 transformations. For example cropping an image after rotation.
                 (Defaults to None)
           image_list_to_transform (list): list of image file paths for images
                                to transform if they are different to the reference
                                images. If None then will use the same image as
                                the reference image. (Defaults to None)
    '''
    def __init__(self, image_list: list, # list of image file names
                       metrics: dict={},
                       metric_images: dict={},
                       image_loader=IQM_Vis.utils.load_image,     # function to load image files
                       image_pre_processing=None,  # apply a function to the image before transformations (e.g. resize to smaller)
                       image_post_processing=None,  # apply a function to the image after transformations (e.g. zoom to help with black boarders on rotation)
                       image_list_to_transform=None, # if you want to use a different image to transform than reference
                       human_exp_csv=None    # csv for where the human experiments file is
                       ):
        self.image_storer = namedtuple('image', ['name', 'data'])
        self.bytes_arrays = namedtuple('arr', ['bytes', 'dtype', 'shape'])
        self.image_loader = image_loader
        self.image_pre_processing = image_pre_processing
        self.load_image_list(image_list)
        if image_list_to_transform != None:
            self.image_list_to_transform = image_list_to_transform
            self._load_image_data(0)   # load the first transform image
        self.metrics = metrics
        for m in self.metrics:
            self.metrics[m] = cache_metric_call(self.metrics[m])
        self.metric_images = metric_images
        self.image_post_processing = image_post_processing
        self.image_post_processing_hash = None
        if human_exp_csv is not None:
            self.human_exp_df = pd.read_csv(human_exp_csv, index_col=0)

        self._check_inputs()

    def add_metric(self, key, value):
        if not isinstance(value, cache_metric_call):
            value = cache_metric_call(value)
        self.metrics[key] = value

    def add_metric_image(self, key, value):
        self.metric_images[key] = value

    def get_image_dataset_list(self):
        # get image file list
        return self.image_list

    def load_image_list(self, image_list):
        if len(image_list) == 0:
            if not hasattr(self, 'image_list'):
                raise ValueError(f'image_list is empty')
            else:
                return
        # remove any non image file paths
        just_images = []
        for image_file in image_list:
            try:
                image_format = imghdr.what(image_file)
            except FileNotFoundError:
                image_format = None
            if image_format != None:
                just_images.append(image_file)
        self.image_list = just_images
        self.image_list_to_transform = just_images
        self.image_names = [get_image_name(file) for file in self.image_list]
        self._load_image_data(0)


    def _load_image_data(self, i):
        # reference image
        self.image_post_processing_hash = None
        self.current_file = self.image_list[i]
        image_name_ref = get_image_name(self.current_file)
        image_data_ref = self._cached_image_loader(self.current_file)
        self.reference_unprocessed = image_data_ref
        if self.image_pre_processing is not None:
            image_data_ref = self.image_pre_processing(image_data_ref)
        self.image_reference = self.image_storer(image_name_ref, image_data_ref)
        self.ref_bytes = self.bytes_arrays(
            image_data_ref.tobytes(), image_data_ref.dtype, image_data_ref.shape)

        # image to transform
        if self.current_file == self.image_list_to_transform[i]:
            self.image_to_transform = self.image_storer(image_name_ref, image_data_ref)
        else:
            image_name_trans = get_image_name(self.image_list_to_transform[i])
            image_data_trans = self._cached_image_loader(
                self.image_list_to_transform[i])
            if self.image_pre_processing is not None:
                image_data_trans = self.image_pre_processing(image_data_trans)
            self.image_reference = self.image_storer(image_name_trans, image_data_trans)

        # Human experiments
        if hasattr(self, 'human_scores'):
            del self.human_scores  # delete old scores (incase we dont have ones for new image)
        if hasattr(self, 'human_exp_df'):
            if image_name_ref in self.human_exp_df.index:
                self.human_scores = {'mean': self.human_exp_df.loc[image_name_ref].to_dict()}

    def __len__(self):
        return len(self.image_list)

    def __getitem__(self, i):
        self._load_image_data(i)

    @cache_tracked
    def _cached_image_loader(self, file_name):
        return self.image_loader(file_name)

    def get_reference_image_by_index(self, index):
        if index >= len(self.image_list):
            raise IndexError('Index out of range of the length of the image list')
        file_name = self.image_list[index]
        image_data = self._cached_image_loader(file_name)
        return image_data

    def get_reference_image_name(self):
        return self.image_reference.name
    
    def get_reference_unprocessed(self):
        return self.reference_unprocessed

    def get_reference_image(self):
        if hash(self.image_post_processing) != self.image_post_processing_hash:
            # need to post process ref image as either first call or post processing has changed
            self.image_reference_post_processed = self.image_reference.data.copy()
            if self.image_post_processing is not None:
                self.image_reference_post_processed = self.image_post_processing(
                    self.image_reference_post_processed)
            # cache the hash so we can test if the post processing changes
            self.image_post_processing_hash = hash(self.image_post_processing)
            # save the bytes array 
            self.ref_bytes = self.bytes_arrays(
                self.image_reference_post_processed.tobytes(), 
                self.image_reference_post_processed.dtype, 
                self.image_reference_post_processed.shape)
        return self.image_reference_post_processed

    def get_image_to_transform_name(self):
        return self.image_to_transform.name

    def get_image_to_transform(self):
        return self.image_to_transform.data

    def get_metrics(self, transformed_image, metrics_to_use='all', **kwargs):
        # convert array to hashable so we can cache already calculated
        trans_bytes = self.bytes_arrays(
            transformed_image.tobytes(), transformed_image.dtype, transformed_image.shape)
        # get metrics
        results = {}
        for metric in self.metrics:
            if metric in metrics_to_use or metrics_to_use == 'all':
                if self.ref_bytes.shape != trans_bytes.shape:
                    # There has been a change in the data so need to quit this calc ASAP
                    results[metric] = 100
                else:
                    # calc as normal
                    results[metric] = self.metrics[metric](
                        self.ref_bytes, trans_bytes, **kwargs)
        return results

    def get_metric_images(self, transformed_image, metrics_to_use='all', **kwargs):
        results = {}
        for metric in self.metric_images:
            if metric in metrics_to_use or metrics_to_use == 'all':
                results[metric] = self.metric_images[metric](self.get_reference_image(), transformed_image, **kwargs)
        return results

    def _check_inputs(self):
        input_types = [(self.image_reference.name, str),
                       (self.image_reference.data, np.ndarray),
                       (self.metrics, dict),
                       (self.metric_images, dict)]
        for item in input_types:
            if type(item[0]) != item[1]:
                var_name = f'{item[0]=}'.split('=')[0]
                raise TypeError(f'holder input: {var_name} should be a {item[1]} not {type(item[0])}')
            
    def clear_all_cache(self):
        for cached_func in CACHED_FUNCTIONS:
            cached_func.cache_clear()


def get_image_name(file_path):
    # get name of image from filepath
    return os.path.splitext(os.path.basename(file_path))[0]


''' thread and signal classes to makethe UI smoother
    useful info about PyQt6 threads: https://www.pythontutorial.net/pyqt/pyqt-qthread/'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from PyQt6.QtCore import QObject, pyqtSignal, pyqtSlot
from IQM_Vis.utils import plot_utils

import time


class get_range_results_worker(QObject):
    progress = pyqtSignal(int)
    current_image = pyqtSignal(str)
    completed = pyqtSignal(dict)
    stopped = pyqtSignal(bool)

    @pyqtSlot(dict)
    def do_work(self, data):
        # t0 = time.time()

        metric_over_range_results = []
        max_val = 0
        self.stop_flag = [False]
        for i, data_store in enumerate(data['data_stores']):
            self.current_image.emit(f'Getting range plot values for {data_store.get_reference_image_name()} progress:')
            results = plot_utils.compute_metrics_over_range_single_trans(data_store,
                data['trans'],
                data['metric_params'],
                data['metrics_to_use'],
                pbar_signal=self.progress,
                stop_flag=self.stop_flag,
                num_steps=data['num_steps'])
            if self.stop_flag[0] == True:
                self.stopped.emit(True)
                return
            metric_over_range_results.append(results)
            # see max metric values
            for _, metric in results.items():
                for trans, data in metric.items():
                    for val in data['scores']:
                        max_val = max(max_val, val)
        data_return = {'metric_over_range_results': metric_over_range_results,
                       'max_val': max_val}
        self.completed.emit(data_return)

        # t1 = time.time()
        # print(f'time: {t1-t0}')

    def stop(self):
        if hasattr(self, 'stop_flag'):
            # use a mutable data type (list) as can pass as a reference (ish) then
            self.stop_flag[0] = True

    def __del__(self):
        # close app upon garbage collection
        self.stop()


'''
image helper functions
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
import PIL
import cv2
import numpy as np
from skimage.transform import resize
from skimage.color import rgb2lab, lab2rgb
from skimage.util import img_as_ubyte

def get_transform_image(data_store, transform_functions, transform_params):
    '''transform image with image post processing

    Args:
        data_store: IQM_Vis data_api
        transform_functions: dict holding transforms
                               (each key is the name of transform, items have key 'function')
        transform_params:  dict holding the parameters for transforms (corresponding to keys in transform_functions)

    Return:
        image: processed numpy image
    '''
    image = data_store.get_image_to_transform()
    for key in transform_functions:
        image = transform_functions[key]['function'](image, transform_params[key])
    if hasattr(data_store, 'image_post_processing'):
        if data_store.image_post_processing is not None:
            image = data_store.image_post_processing(image)
    return image

def load_image(image_path):
    '''
    load image as RGB float
    '''
    if not os.path.isfile(image_path):
        raise ValueError(f'Image file: {image_path} does not exist')
    img = PIL.Image.open(image_path)
    img = np.array(img)
    if img.dtype == 'uint8':
        img = img/255
    if img.shape[2] > 3:
        img = img[:, :, :3]
    return img

def save_image(img, path):
    ''' save image as ubyte '''
    img = PIL.Image.fromarray(img_as_ubyte(img))
    img.save(path)

def resize_to_longest_side(im, side=128):
    '''
    resize image to longest side
    '''
    shape = im.shape
    if shape[0] > shape[1]:
        scale = side/shape[0]
        size = (side, int(shape[1]*scale))
    else:
        scale = side/shape[1]
        size = (int(shape[0]*scale), side)
    im = resize_image(im, size)
    return im

def resize_image(img, size=128):
    '''
    resize image to square or specified size
    '''
    if isinstance(size, int):
        size = (size, size)
    img = cv2.resize(img, (size[1], size[0]), interpolation=cv2.INTER_LINEAR)  # cv2 much faster than skimage

    # can change to PIL code below but only works with uint8 images
    # img = PIL.Image.fromarray(img_as_ubyte(img))
    # img = img.resize(size)
    # img = np.array(img)/255

    return img

def crop_centre(image, scale_factor=2, keep_size=True):
    ''' crop to the centre of the image, note this will return a small image size
        so it best used as post processing

    Args:
        image (np.array): image to be cropped
        scale_factor (float): the percentage to zoom in by (for square crop only).
                              0.5  = 2x zoom out
                              1 = normal size
                              2 = 2x zoom in
                              (Defaults to 2 - half the size)
        keep_size (bool): resize image to the original size after cropping (Defaults to True)

    Returns:
        image (np.array): cropeed image
    '''
    original_size = image.shape
    new_size_x = int(original_size[0]/scale_factor)
    new_size_y = int(original_size[1]/scale_factor)
    if scale_factor > 1:
        start_point_x = int((original_size[0] - new_size_x)/2)
        start_point_y = int((original_size[1] - new_size_y)/2)
        image = image[start_point_x:start_point_x+new_size_x,
                      start_point_y:start_point_y+new_size_y,
                      :]
    elif scale_factor < 1:
        new_size = (new_size_x, new_size_y, original_size[2]) if len(original_size) == 3 else (new_size_x, new_size_y)
        start_point_x = int((new_size_x - original_size[0])/2)
        start_point_y = int((new_size_y - original_size[1])/2)
        zoomed_out = np.zeros(new_size, dtype=image.dtype)
        zoomed_out[start_point_x:start_point_x+original_size[0],
                   start_point_y:start_point_y+original_size[1],
                   :] = image
        image = zoomed_out
    
    if keep_size == True:
        image = resize_image(image, (original_size[0], original_size[1]))

    return image

def calibrate_brightness(im, rgb_brightness, display_brightness, ubyte=True):
    if rgb_brightness == display_brightness:
        return im
    if ubyte == True:
        im = im/255
    im_lab = rgb2lab(im)
    scale = rgb_brightness/display_brightness
    scaled_brightness = im_lab[:, :, 0]*scale
    clipped = np.clip(scaled_brightness, 0, 100)
    im_lab[:, :, 0] = clipped
    rgb = lab2rgb(im_lab)
    if ubyte == True:
        rgb = img_as_ubyte(rgb)
    return rgb



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from IQM_Vis.utils.image_utils import *
from IQM_Vis.utils.plot_utils import bar_plotter, radar_plotter, compute_metrics_over_range_single_trans, get_radar_plots_avg_plots, get_transform_range_plots
from IQM_Vis.utils.save_utils import *


'''
create experiment window JND
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
import random
import threading
import warnings
import time
from functools import partial
import copy

import numpy as np
import pandas as pd
from PyQt6.QtWidgets import (QMainWindow,
                             QHBoxLayout,
                             QVBoxLayout,
                             QTabWidget,
                             QApplication,
                             QPushButton,
                             QLabel,
                             QMessageBox)

from PyQt6.QtCore import Qt, pyqtSignal, pyqtSlot, QObject, QThread
from PyQt6.QtGui import QShortcut, QKeySequence

import IQM_Vis
from IQM_Vis.UI.custom_widgets import ClickLabel
from IQM_Vis.UI import utils
from IQM_Vis.utils import gui_utils, plot_utils, image_utils, save_utils


class make_experiment_JND(QMainWindow):
    '''https://www.verywellmind.com/what-is-the-just-noticeable-difference-2795306'''
    saved_experiment = pyqtSignal(str)
    reset_clicked_image = pyqtSignal(dict)

    def __init__(self, 
                 checked_transformation_params, 
                 data_store, 
                 image_display_size,
                 rgb_brightness,
                 display_brightness,
                 default_save_dir=save_utils.DEFAULT_SAVE_DIR,
                 dataset_name='dataset1',
                 image_preprocessing='None',
                 image_postprocessing='None',
                 lower_im_num=1,
                 upper_im_num=1,
                 checked_metrics={}):
        super().__init__()
        self.checked_transformation_params = checked_transformation_params
        if self.checked_transformation_params == {}:
            return
        elif len(self.checked_transformation_params) != 1:
            raise AttributeError(f'Just Noticable difference experiment can only use one transform/distortion')

        self.checked_metrics = checked_metrics
        self.data_store = copy.copy(data_store)
        self.image_display_size = image_display_size
        self.rgb_brightness = rgb_brightness
        self.display_brightness = display_brightness
        self.default_save_dir = os.path.join(default_save_dir, 'JND')
        self.dataset_name = dataset_name
        self.default_save_dir = os.path.join(
            self.default_save_dir, self.dataset_name)
        self.dataset_name = dataset_name
        self.curr_im_ind = 0
        self.save_im_format = '.png'
        self.lower_im_num = lower_im_num
        self.upper_im_num = upper_im_num

        self.processing = {'pre': image_preprocessing,
                           'post': image_postprocessing}

        # self.image_change_worker = reset_image_widget_to_black()
        # self.image_change_worker.completed.connect(self.click_completed)
        # self.image_worker_thread = QThread()
        # self.reset_clicked_image.connect(self.image_change_worker.change_to_solid)
        # self.image_change_worker.moveToThread(self.image_worker_thread)
        # self.image_worker_thread.start()

        self.stop_event = threading.Event()
        self.saved = False
        self.quit_experiment = False
        self._init_experiment_window_widgets()
        self.experiment_layout()
        self.setCentralWidget(self.experiments_tab)
        self.setWindowTitle('JND Experiment')
        # move to centre of the screen
        qr = self.frameGeometry()
        cp = self.screen().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())
        # wait for the window to show before loading images
        # self.show()
        QApplication.processEvents()
        # get all images and show them
        self.get_all_images()

        # make unique save dir name
        self.new_save_dir = self.get_unique_save_dir()

        self.show_all_images()
        self.get_metric_scores()

    def closeEvent(self, event):
        # Ask for confirmation if not saved
        if not self.saved:
            answer = QMessageBox.question(self,
            "Confirm Exit...",
            "Are you sure you want to exit?\nAll unsaved data will be lost.",
            QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes, 
                                          QMessageBox.StandardButton.Yes)
        else:
            answer = QMessageBox.StandardButton.Yes

        event.ignore()
        if answer == QMessageBox.StandardButton.Yes:
            self.quit_experiment = True
            if hasattr(self, 'range_worker'):
                self.range_worker.stop()
            if hasattr(self, 'image_worker_thread'):
                if self.image_worker_thread.isRunning():
                    self.image_worker_thread.quit()
                    self.image_worker_thread.wait()
                self.image_change_worker.stop()
            self.stop_event.set()
            # self.clicked_event.set()
            event.accept()

    def quit(self):
        self.close()

    def show_all_images(self, tab='setup'):
        self.widget_experiments['setup']['text'].setText(f'''
        JND Experiment to be setup with the above images using the settings:
            Save folder: {self.default_save_dir}
            Image Display Size: {self.image_display_size}
            Image Calibration:
                Max RGB Brightness: {self.rgb_brightness}
                Max Display Brightness: {self.display_brightness}

            Number of Comparisons: {int(len(self.experiment_transforms))}

        Click the Setup button to setup up the experiment and hand over to the test subject.
        ''')
        self.widget_experiments[tab]['images'].axes.axis('off')
        rows = min(int(len(self.experiment_transforms)**0.5), 5)
        cols = min(int(np.ceil(len(self.experiment_transforms)/rows)), 5)
        for i, trans in enumerate(self.experiment_transforms):
            if i == rows*cols:
                break
            ax = self.widget_experiments[tab]['images'].figure.add_subplot(
                rows, cols, i+1)
            ax.imshow(image_utils.calibrate_brightness(
                trans['image'], self.rgb_brightness, self.display_brightness, ubyte=False))
            if tab == 'final':
                ax.set_ylabel('')
                ax.set_xlabel(trans['user_decision'], fontsize=6)
                ax.set_xticks([])
                ax.set_yticks([])
            else:
                ax.axis('off')
            ax.set_title(save_utils.make_name_for_trans(trans), fontsize=6)
        # self.widget_experiments[tab]['images'].figure.tight_layout()

        # time.sleep(5)
        # QApplication.processEvents()
            
    def get_all_images(self):
        
        # get all the transform values
        self.experiment_trans_params = plot_utils.get_all_single_transform_params(
            self.checked_transformation_params, num_steps='from_dict')

        # save the experiment ordering before reordering (for saving to csv col ordering)
        self.original_params_order = []
        for single_trans in self.experiment_trans_params:
            trans_name = list(single_trans.keys())[0]
            param = single_trans[trans_name]
            data = {'transform_name': trans_name,
                    'transform_value': param}
            self.original_params_order.append(
                save_utils.make_name_for_trans(data))


        # all images in the dataset
        self.all_ref_images = {}
        self.experiment_transforms = []

        for i in range(max(self.lower_im_num-1, 0), min(len(self.data_store), self.upper_im_num)):
            # load the image in dataset
            self.data_store[i]
            # REFERENCE image
            ref_image = self.data_store.get_reference_image()
            ref_name = self.data_store.get_reference_image_name()
            # save name and image
            self.all_ref_images[ref_name] = ref_image
            # if hasattr(self.data_store, 'get_reference_unprocessed'):
            #     self.ref_image_unprocessed = self.data_store.get_reference_unprocessed()
            # get all transformed images
            for single_trans in self.experiment_trans_params:
                trans_name = list(single_trans.keys())[0]
                param = single_trans[trans_name]
                img = self.get_single_transform_im(single_trans)
                data = {'transform_name': trans_name,
                        'transform_value': param,
                        'image': img,
                        'ref_name': ref_name}
                self.experiment_transforms.append(data)
        
            # add reference image to list
            data = {'transform_name': 'None',
                    'transform_value': 'None',
                    'image': ref_image,
                    'ref_name': ref_name}
            self.experiment_transforms.append(data)

            # update user
            self.widget_experiments['setup']['text'].setText(
                f'''Loading all images in the dataset {i+1}/{len(self.data_store)}''')

        # shuffle the images list
        random.shuffle(self.experiment_transforms)

    def get_metric_scores(self, calc=False):
        '''get IQM scores to save alongside the experiment for plotting/analysis purposes'''
        self.IQM_scores_df = None
        if len(self.checked_metrics) == 0:
            return
        if calc == True:
            ''' this currently doesn't work and might be a bit taxing to do for all images'''
            IQM_scores = {}
            for data in self.experiment_transforms:
                score_dict = self.data_store.get_metrics(transformed_image=data['image'],
                                                    metrics_to_use=self.checked_metrics)
                scores = []
                metrics = []
                for name, score in score_dict.items():
                    metrics.append(name)
                    scores.append(float(score))
                IQM_scores[save_utils.make_name_for_trans(data)] = scores
            IQM_scores['IQM'] = metrics
            self.IQM_scores_df = pd.DataFrame.from_dict(IQM_scores)
            self.IQM_scores_df.set_index('IQM', inplace=True)

    def _init_experiment_window_widgets(self):
        self.widget_experiments = {'exp': {}, 'preamble': {}, 'setup': {}, 'final':{}}
        ''' setup tab '''
        self.widget_experiments['setup']['start_button'] = QPushButton(
            'Setup', self)
        self.widget_experiments['setup']['start_button'].clicked.connect(self.setup_experiment)
        self.widget_experiments['setup']['quit_button'] = QPushButton('Quit', self)
        self.widget_experiments['setup']['quit_button'].clicked.connect(self.quit)
        QShortcut(QKeySequence("Ctrl+Q"),
                  self.widget_experiments['setup']['quit_button'], self.quit)
        self.widget_experiments['setup']['images'] = gui_utils.MplCanvas(size=None)
        self.widget_experiments['setup']['text'] = QLabel(self)
        self.widget_experiments['setup']['text'].setText(f'''Loading all images in the dataset''')
        # self.widget_experiments['setup']['text'].setAlignment(
        #     Qt.AlignmentFlag.AlignCenter)

        ''' info tab '''
        self.widget_experiments['preamble']['text'] = QLabel(self)
        self.widget_experiments['preamble']['text'].setText('''
        For this experiment you will be shown a reference image a comparison image.

        You need to click SAME or DIFFERENT whether you think the comparison image is the same or different to the reference image.




        When you are ready, click the Start button to begin the experiment ''')
        self.running_experiment = False
        self.widget_experiments['preamble']['start_button'] = QPushButton('Start', self)
        self.widget_experiments['preamble']['start_button'].clicked.connect(self.toggle_experiment)

        self.widget_experiments['preamble']['quit_button'] = QPushButton('Quit', self)
        self.widget_experiments['preamble']['quit_button'].clicked.connect(
            self.quit)
        QShortcut(QKeySequence("Ctrl+Q"),
                  self.widget_experiments['preamble']['quit_button'], self.quit)

        ''' experiment tab '''
        self.exp_info_text = 'Click same or different for the two images shown (or press the S or D key)'
        self.widget_experiments['exp']['info'] = QLabel(self.exp_info_text, self)
        for image in ['Reference', 'Comparison']:
            self.widget_experiments['exp'][image] = {}
            self.widget_experiments['exp'][image]['data'] = ClickLabel(image)
            self.widget_experiments['exp'][image]['data'].setAlignment(Qt.AlignmentFlag.AlignCenter)
            # image label
            self.widget_experiments['exp'][image]['label'] = QLabel(image, self)
            self.widget_experiments['exp'][image]['label'].setAlignment(Qt.AlignmentFlag.AlignCenter)
        
        self.widget_experiments['exp']['same_button'] = QPushButton('SAME', self)
        self.widget_experiments['exp']['same_button'].clicked.connect(partial(self.user_decision, 'same'))
        self.widget_experiments['exp']['diff_button'] = QPushButton('DIFFERENT', self)
        self.widget_experiments['exp']['diff_button'].clicked.connect(partial(self.user_decision, 'diff'))
        self.widget_experiments['exp']['quit_button'] = QPushButton('Quit', self)
        self.widget_experiments['exp']['quit_button'].clicked.connect(self.quit)
        QShortcut(QKeySequence("S"), self.widget_experiments['exp']['same_button'], partial(
            self.user_decision, 'same'))
        QShortcut(QKeySequence("D"), self.widget_experiments['exp']['same_button'], partial(
            self.user_decision, 'diff'))
        QShortcut(QKeySequence("Ctrl+Q"),
                  self.widget_experiments['exp']['quit_button'], self.quit)

        ''' finish tab '''
        self.widget_experiments['final']['order_text'] = QLabel(
            'Experiment Results:', self)
        self.widget_experiments['final']['images'] = gui_utils.MplCanvas(size=None)
        self.widget_experiments['final']['quit_button'] = QPushButton('Quit', self)
        self.widget_experiments['final']['quit_button'].clicked.connect(
            self.quit)
        QShortcut(QKeySequence("Ctrl+Q"),
                self.widget_experiments['final']['quit_button'], self.quit)
        self.widget_experiments['final']['save_label'] = QLabel('Not saved yet', self)

    def experiment_layout(self):
        ''' setup '''
        experiment_text = QVBoxLayout()
        experiment_text.addWidget(self.widget_experiments['setup']['text'])
        experiment_setup_buttons = QHBoxLayout()
        experiment_setup_buttons.addWidget(
            self.widget_experiments['setup']['start_button'])
        experiment_setup_buttons.addWidget(
            self.widget_experiments['setup']['quit_button'])
        experiment_text.addLayout(experiment_setup_buttons)

        experiment_mode_setup = QVBoxLayout()
        experiment_mode_setup.addWidget(self.widget_experiments['setup']['images'])
        experiment_mode_setup.addLayout(experiment_text)
        experiment_mode_setup.setAlignment(Qt.AlignmentFlag.AlignCenter)
        experiment_mode_setup.addStretch()


        ''' info '''
        experiment_info_buttons = QHBoxLayout()
        experiment_info_buttons.addWidget(
            self.widget_experiments['preamble']['start_button'])
        experiment_info_buttons.addWidget(
            self.widget_experiments['preamble']['quit_button'])

        experiment_mode_info = QVBoxLayout()
        experiment_mode_info.addWidget(
            self.widget_experiments['preamble']['text'])
        experiment_mode_info.setAlignment(Qt.AlignmentFlag.AlignCenter)
        experiment_mode_info.addLayout(experiment_info_buttons)

        ''' experiment '''
        same_diff_button = QHBoxLayout()
        same_diff_button.addWidget(self.widget_experiments['exp']['same_button'])
        same_diff_button.addWidget(self.widget_experiments['exp']['diff_button'])
        same_diff_button.setAlignment(Qt.AlignmentFlag.AlignCenter)

        info = QVBoxLayout()
        info.addWidget(self.widget_experiments['exp']['info'])
        info.setAlignment(Qt.AlignmentFlag.AlignCenter)

        quit_button = QVBoxLayout()
        quit_button.addWidget(self.widget_experiments['exp']['quit_button'])
        quit_button.setAlignment(Qt.AlignmentFlag.AlignCenter)

        layouts = []
        for im in ['Reference', 'Comparison']:
            _layout = QVBoxLayout()
            for _, widget in self.widget_experiments['exp'][im].items():
                _layout.addWidget(widget)
            _layout.setAlignment(Qt.AlignmentFlag.AlignTop)
            layouts.append(_layout)

        # add images to h box
        experiment_images = QHBoxLayout()
        for layout in layouts:
            experiment_images.addLayout(layout)
        experiment_images.setAlignment(Qt.AlignmentFlag.AlignTop)

        run_experiment = QVBoxLayout()
        run_experiment.addLayout(info)
        run_experiment.addLayout(same_diff_button)
        run_experiment.addLayout(experiment_images)
        run_experiment.addLayout(quit_button)
        run_experiment.setAlignment(Qt.AlignmentFlag.AlignCenter)

        ''' finished '''
        finish_experiment = QVBoxLayout()
        finish_experiment.addWidget(self.widget_experiments['final']['order_text'])
        finish_experiment.addWidget(self.widget_experiments['final']['images'])
        finish_experiment.addWidget(self.widget_experiments['final']['save_label'])
        finish_experiment.addWidget(self.widget_experiments['final']['quit_button'])
        finish_experiment.setAlignment(Qt.AlignmentFlag.AlignCenter)
        finish_experiment.addStretch()

        self.experiments_tab = QTabWidget()
        for tab_layout, tab_name in zip([experiment_mode_setup, experiment_mode_info, run_experiment, finish_experiment],
                                        ['setup', 'info', 'run', 'finish']):
            utils.add_layout_to_tab(self.experiments_tab, tab_layout, tab_name)
        # experiment_mode_layout = QVBoxLayout()
        # experiment_mode_layout.addWidget(self.experiments_tab)
        # return experiment_mode_layout

    ''' experiment running functions'''
    def setup_experiment(self):
        self.experiments_tab.setCurrentIndex(1)
        self.experiments_tab.setTabEnabled(0, False)
        self.experiments_tab.setTabEnabled(2, False)
        self.experiments_tab.setTabEnabled(3, False)

    def toggle_experiment(self):
        if self.running_experiment:
            self.reset_experiment()
            self.experiments_tab.setTabEnabled(0, True)
            self.experiments_tab.setTabEnabled(1, True)
            self.running_experiment = False
            # self.widget_experiments['preamble']['start_button'].setText('Start')

        else:
            self.experiments_tab.setTabEnabled(2, True)
            self.start_experiment()
            self.experiments_tab.setTabEnabled(0, False)
            self.experiments_tab.setTabEnabled(1, False)
            # self.widget_experiments['preamble']['start_button'].setText('Reset')
            self.running_experiment = True

    def reset_experiment(self):
        self.experiments_tab.setCurrentIndex(1)
        self.init_style('light')

    def start_experiment(self):
        self.init_style('dark')
        self.experiments_tab.setCurrentIndex(2)

        # Display reference image
        gui_utils.change_im(self.widget_experiments['exp']['Reference']['data'], 
                            self.all_ref_images[self.experiment_transforms[self.curr_im_ind]['ref_name']],
                            resize=self.image_display_size, 
                            rgb_brightness=self.rgb_brightness, 
                            display_brightness=self.display_brightness)

        # exp data holder
        self.time0 = time.time()
        self.curr_im_ind = 0
        # Display comparison image
        gui_utils.change_im(self.widget_experiments['exp']['Comparison']['data'], 
                            self.experiment_transforms[self.curr_im_ind]['image'],
                            resize=self.image_display_size, 
                            rgb_brightness=self.rgb_brightness, 
                            display_brightness=self.display_brightness)

    def finish_experiment(self):
        self.experiments_tab.setTabEnabled(3, True)
        self.show_all_images(tab='final')
        self.init_style('light')
        self.experiments_tab.setCurrentIndex(3)
        # self.experiments_tab.setTabEnabled(2, False)

        # save experiment to file
        self.save_experiment()
        if self.saved == True:
            self.widget_experiments['final']['save_label'].setText(f'Saved to {self.default_save_dir}')
        else:
            self.widget_experiments['final']['save_label'].setText(f'Save failed to {self.default_save_dir}')

    def get_trans_funcs(self):
        # get the current transform functions
        trans_funcs = {}
        for single_trans in self.experiment_trans_params:
            trans_name = list(single_trans.keys())[0]
            trans_funcs[trans_name] = self.checked_transformation_params[trans_name]['function']
        return trans_funcs
    
    def get_unique_save_dir(self):
        '''get directory that is unique based on if it's the same experiment or not'''
        trans_funcs = self.get_trans_funcs()

        # get a unique directory (same image with diff trans need a new dir)
        i = 1
        unique_dir_found = False
        new_dir = True
        while unique_dir_found == False:
            exp_save_dir = os.path.join(self.default_save_dir, f'experiment-{i}')
            if os.path.exists(exp_save_dir):
                # get transform funcs and params
                exp_trans_params = save_utils.load_obj(
                    os.path.join(exp_save_dir, 'transforms', 'transform_params.pkl'))
                exp_trans_funcs = save_utils.load_obj(
                    os.path.join(exp_save_dir, 'transforms', 'transform_functions.pkl'))
                
                # get ref image names
                im_names = save_utils.get_JND_image_names(exp_save_dir)
                im_names.sort()
                curr_im_names = list(self.all_ref_images.keys())
                # add file extension to names
                curr_im_names = [f'{name}{self.save_im_format}' for name in curr_im_names]
                curr_im_names.sort()

                # get image processing saved params
                processing_file = save_utils.get_image_processing_file(
                    exp_save_dir)
                procesing_same = False
                if os.path.exists(processing_file):
                    processing = save_utils.load_json_dict(processing_file)
                    if processing == self.processing:
                        procesing_same = True

                # check if experiment is the same
                if ((exp_trans_params == self.original_params_order) 
                    and (trans_funcs == exp_trans_funcs) 
                    and procesing_same
                    and (im_names == curr_im_names)):
                    self.default_save_dir = exp_save_dir
                    unique_dir_found = True
                    new_dir = False
                else:
                    i += 1
            else:
                self.default_save_dir = exp_save_dir
                unique_dir_found = True

        return new_dir

    def save_experiment(self):
        # get the current transform functions
        trans_funcs = self.get_trans_funcs()
        
        # make all the dirs and subdirs
        os.makedirs(self.default_save_dir, exist_ok=True)
        os.makedirs(os.path.join(save_utils.get_JND_ref_image_dir(self.default_save_dir)), exist_ok=True)
        os.makedirs(os.path.join(self.default_save_dir, 'images'), exist_ok=True)
        os.makedirs(os.path.join(self.default_save_dir, 'transforms'), exist_ok=True)

        if not os.path.exists(save_utils.get_JND_ref_image_unprocessed_dir(self.default_save_dir)):
            if hasattr(self, 'ref_image_unprocessed'):
                image_utils.save_image(self.ref_image_unprocessed,
                                        save_utils.get_original_unprocessed_image_file(self.default_save_dir))
        if self.new_save_dir == True:
            # save experiment images
            for name, im in self.all_ref_images.items():
                image_utils.save_image(im,
                                       os.path.join(save_utils.get_JND_ref_image_dir(self.default_save_dir), 
                                                    f'{name}{self.save_im_format}'))
            for trans in self.experiment_transforms:
                image_utils.save_image(
                    trans['image'], 
                    os.path.join(self.default_save_dir, 
                                 'images',
                                 f"{save_utils.make_name_for_trans(trans)}-{trans['ref_name']}{self.save_im_format}",
                                 ))
            # save the transformations
            save_utils.save_obj(
                save_utils.get_transform_params_file(self.default_save_dir),
                self.original_params_order)
            save_utils.save_obj(
                save_utils.get_transform_functions_file(self.default_save_dir),
                dict(sorted(trans_funcs.items())))
            # save the image pre/post processing options
            save_utils.save_json_dict(
                save_utils.get_image_processing_file(self.default_save_dir),
                self.processing)

        # save the experiment results
        csv_file = save_utils.save_JND_experiment_results(
            self.experiment_transforms,
            self.default_save_dir,
            self.IQM_scores_df)
        self.saved = True
        self.saved_experiment.emit(csv_file)

    def user_decision(self, decision):
        if decision not in ['same', 'diff']:
            raise ValueError(f'user decision for JND experiment needs to be same or diff')
        # make sure we don't go beyond the data set with acciental key presses
        if self.curr_im_ind >=  len(self.experiment_transforms):
            return
        # save time it took 

        # log decision
        self.experiment_transforms[self.curr_im_ind]['user_decision'] = decision
        self.experiment_transforms[self.curr_im_ind]['time_taken'] = time.time()-self.time0

        # move to next image
        self.curr_im_ind += 1
        if self.curr_im_ind ==  len(self.experiment_transforms):
            self.finish_experiment()
        else:
            # Display reference image
            gui_utils.change_im(self.widget_experiments['exp']['Reference']['data'],
                                self.all_ref_images[self.experiment_transforms[self.curr_im_ind]['ref_name']],
                                resize=self.image_display_size,
                                rgb_brightness=self.rgb_brightness,
                                display_brightness=self.display_brightness)
            # Comparison image
            gui_utils.change_im(self.widget_experiments['exp']['Comparison']['data'], 
                                self.experiment_transforms[self.curr_im_ind]['image'],
                                resize=self.image_display_size, 
                                rgb_brightness=self.rgb_brightness, 
                                display_brightness=self.display_brightness)
            self.widget_experiments['exp']['info'].setText(
                f'{self.exp_info_text} {self.curr_im_ind+1}/{len(self.experiment_transforms)}')

        # reset time
        self.time0 = time.time()

    def get_single_transform_im(self, single_trans_dict):
        trans_name = list(single_trans_dict)[0]
        return image_utils.get_transform_image(self.data_store,
                                        {trans_name: self.checked_transformation_params[trans_name]},
                                        single_trans_dict)

    ''' UI '''
    def init_style(self, style='light', css_file=None):
        if css_file == None:
            dir = os.path.dirname(os.path.abspath(__file__))
            # css_file = os.path.join(dir, 'style-light.css')
            css_file = os.path.join(dir, f'style-{style}.css')
        if os.path.isfile(css_file):
            with open(css_file, 'r') as file:
                self.setStyleSheet(file.read())
        else:
            warnings.warn('Cannot load css style sheet - file not found')


class reset_image_widget_to_black(QObject):
    ''' change clicked image to black and pause '''
    completed = pyqtSignal(float)

    def __init__(self, time=0.1):
        super().__init__()
        self.running = True
        self.time = time
        self.black_array = np.zeros([100, 100, 3])
        # self.black_array[:, :, 1] = 1 # blue

    @pyqtSlot(dict)
    def change_to_solid(self, data):
        t_start = time.time()
        image_display_size = data['image_display_size']
        widget = data['widget']
        # make clicked image black to show user
        gui_utils.change_im(widget,
                            self.black_array,
                            resize=image_display_size)
        # pause for half the time needed (will use the loop below to wait for full time and also see if image has turned black yet)
        time.sleep(self.time/2)
        m = 1
        t_time = time.time() - t_start
        # wait until black image is shown
        while (not (m == 0 or m == 63.75)) and (t_time < self.time): # 63.75 is when 4th channel is all ones , rest are 0
            time.sleep(self.time/10)
            # get the image data from the widget to check if it's been made black yet
            pixmap = widget.pixmap()
            q_img = pixmap.toImage()
            ptr = q_img.bits()
            ptr = q_img.constBits()
            ptr.setsize(q_img.sizeInBytes())
            np_img = np.array(ptr, copy=False).reshape(
                q_img.height(), q_img.width(), 4)
            m = np_img.mean()
            # calc if the time is up yet
            t_time = time.time() - t_start
        # all complete so send signal 
        self.completed.emit(1.0)

    def stop(self):
        self.running = False

    def __del__(self):
        # close app upon garbage collection
        self.stop()



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

'''
Test that all of the transforms work as expected
'''
import numpy as np
from IQM_Vis.transforms import get_all_transforms
from IQM_Vis.utils import load_image
from IQM_Vis.examples.images import IMAGE2


def _get_test_image():
    return load_image(IMAGE2)


def test_all_transforms_init_value_no_change_image():
    # make a mock image
    image = _get_test_image()
    # test that the default value doesn't change the image
    tol = 0.0
    all_transfoms = get_all_transforms()
    for name, transform in all_transfoms.items():
        if 'init_value' in transform:
            value = transform['init_value']
        else:
            value = 0
        assert np.allclose(transform['function'](image, value), image, rtol=tol)


def test_all_transforms_min_max_specified():
    # make a mock image
    image = _get_test_image()
    # test that the default value doesn't change the image
    tol = 0.0
    all_transfoms = get_all_transforms()
    # test that the min and max values work
    for name, transform in all_transfoms.items():
        assert 'min' in transform or 'max' in transform


def test_all_transforms_min_change_image():
    # make a mock image
    image = _get_test_image()
    # test that the default value doesn't change the image
    tol = 0.0
    all_transfoms = get_all_transforms()
    # test that the min and max values work
    for name, transform in all_transfoms.items():
        if 'init_value' in transform:
            if transform['init_value'] == transform['min']:
                continue
        else:
            if transform['min'] == 0:
                continue
        assert not np.allclose(transform['function'](image, transform['min']), image, rtol=tol)


def test_all_transforms_max_change_image():
    # make a mock image
    image = _get_test_image()
    # test that the default value doesn't change the image
    tol = 0.0
    all_transfoms = get_all_transforms()
    # test that the min and max values work
    for name, transform in all_transfoms.items():
        if 'init_value' in transform:
            if transform['init_value'] == transform['max']:
                continue
        else:
            if transform['max'] == 0:
                continue
        assert not np.allclose(transform['function'](image, transform['max']), image, rtol=tol)


if __name__ == '__main__':
    test_all_transforms_max_change_image()



''' Custom SubClasses of PyQt6 widget to extent to the required functionality '''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import QProgressBar, QLabel


class ClickLabel(QLabel):
    ''' add a click mousePressEvent event to label (so can click on images) '''
    clicked = pyqtSignal(str, str)

    def __init__(self, name='clickable label'):
        super().__init__()
        self.name = name

    def mousePressEvent(self, event=None):
        self.clicked.emit(self.objectName(), self.name)


StyleSheet = '''
#RedProgressBar {
    text-align: center;
}
#RedProgressBar::chunk {
    background-color: #F44336;
}
#GreenProgressBar {
    min-height: 12px;
    max-height: 12px;
    border-radius: 6px;
}
#GreenProgressBar::chunk {
    border-radius: 6px;
    background-color: #009688;
}
#BlueProgressBar {
    border: 2px solid #2196F3;
    border-radius: 5px;
    background-color: #E0E0E0;
}
#BlueProgressBar::chunk {
    background-color: #2196F3;
    width: 10px;
    margin: 0.5px;
}
'''


class ProgressBar(QProgressBar):
    def __init__(self, *args, **kwargs):
        super(ProgressBar, self).__init__(*args, **kwargs)
        self.setValue(0)
        self.setGeometry(100, 50, 200, 30)
        # changing the color of process bar
        self.setStyleSheet("""QProgressBar {
            min-height: 12px;
            max-height: 12px;
            border-radius: 6px;
        }
        QProgressBar::chunk {
            border-radius: 6px;
            background-color: #009688;
        }""")



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License
import warnings
import torch 

from torchmetrics import StructuralSimilarityIndexMeasure as ssim_torch
from IQM_Vis.metrics.SSIM.ssim import ms_ssim
from IQM_Vis.metrics.NLPD_torch.pyramids import LaplacianPyramid
from IQM_Vis.metrics.metric_utils import _check_shapes, _numpy_to_torch_image


class SSIM:
    '''Structural Similarity Index Measure between two images. Images must have
       the same dimensions. Score given is 1 - SSIM to give the loss/dissimilarity

    Args:
        return_image (bool): Whether to return the image (Defaults to False which
                             will return a scalar value)
    '''

    def __init__(self, return_image=False):
        self.return_image = return_image
        self.metric = ssim_torch
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu")
        self.preproccess_function = _numpy_to_torch_image

    def __call__(self, im_ref, im_comp, sigma=1.5, k1=0.01, k2=0.03, ssim_kernel_size=11, **kwargs):
        '''When an instance is called

        Args:
            im_ref (np.array): Reference image
            im_comp (np.array): Comparison image
            **kwargs: Arbitrary keyword arguments

        Returns:
            score (np.array): 1-SSIM (scalar if return_image is False, image if
                             return_image is True)
        '''
        _check_shapes(im_ref, im_comp)
        im_ref = self.preproccess_function(im_ref).to(
            device=self.device, dtype=torch.float)
        im_comp = self.preproccess_function(im_comp).to(
            device=self.device, dtype=torch.float)
        ssim_kernel_size = _make_kernel_odd(ssim_kernel_size)
        # set up metric
        with warnings.catch_warnings():    # we don't care about the warnings these give
            warnings.simplefilter("ignore")
            if self.return_image:
                _metric = self.metric(
                    sigma=sigma, k1=k1, k2=k2, return_full_image=True, reduction=None, kernel_size=ssim_kernel_size)
            else:
                _metric = self.metric(sigma=sigma, k1=k1,
                                      k2=k2, kernel_size=ssim_kernel_size)
            _metric.to(self.device)
        if self.return_image:
            _, ssim_full_im = _metric(im_ref, im_comp)
            ssim_full_im = torch.squeeze(ssim_full_im, axis=0)
            ssim_full_im = ssim_full_im.permute(1, 2, 0)
            ssim_full_im = torch.clip(ssim_full_im, 0, 1)
            _score = ssim_full_im.cpu().detach().numpy()
        else:
            _score = _metric(im_ref, im_comp).cpu().detach().numpy()
        _score = 1 - _score
        _metric.reset()
        return _score


class MS_SSIM:
    '''Multi-Scale Structural Similarity Index Measure between two images.
       Images must have the same dimensions. Score given is 1 - MS_SSIM to give the
       loss/dissimilarity.
       Note that images of small size, below 180 pixels will have their kernel size
       reduced for compatability with the 4 downsizing operations.

    Args:
        return_image (bool): Whether to return the image (Defaults to False which
                             will return a scalar value)
    '''

    def __init__(self, return_image=False):
        self.return_image = return_image
        # self.metric = Mssim_torch
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu")
        self.preproccess_function = _numpy_to_torch_image

    def __call__(self, im_ref, im_comp, sigma=1.5, k1=0.01, k2=0.03, mssim_kernel_size=11, ** kwargs):
        '''When an instance is called

        Args:
            im_ref (np.array): Reference image
            im_comp (np.array): Comparison image
            **kwargs: Arbitrary keyword arguments

        Returns:
            score (np.array): 1-SSIM (scalar if return_image is False, image if
                             return_image is True)
        '''
        _check_shapes(im_ref, im_comp)
        im_ref = self.preproccess_function(im_ref).to(
            device=self.device, dtype=torch.float)
        im_comp = self.preproccess_function(im_comp).to(
            device=self.device, dtype=torch.float)
        mssim_kernel_size = _make_kernel_odd(mssim_kernel_size)
        success = False
        reduced_kernel = False
        run_error = False
        while success == False and mssim_kernel_size > 0:
            # _metric = self._make_metric(sigma=sigma, k1=k1, k2=k2, kernel_size=mssim_kernel_size)
            try:
                if self.return_image:
                    # _, ssim_full_im = _metric(im_ref, im_comp)
                    ssim_full_im = ms_ssim(im_ref, im_comp,
                                           data_range=1,
                                           win_size=mssim_kernel_size,
                                           win_sigma=sigma,
                                           K=(k1, k2),
                                           return_image=True)
                    ssim_full_im = torch.squeeze(ssim_full_im, axis=0)
                    ssim_full_im = ssim_full_im.permute(1, 2, 0)
                    ssim_full_im = torch.clip(ssim_full_im, 0, 1)
                    _score = ssim_full_im.cpu().detach().numpy()
                else:
                    _score = ms_ssim(im_ref, im_comp,
                                     data_range=1,
                                     win_size=mssim_kernel_size,
                                     win_sigma=sigma,
                                     K=(k1, k2),
                                     size_average=True)
                    # _score = _metric(im_ref, im_comp).cpu().detach().numpy()
                _score = 1 - _score.cpu().detach().numpy()  # get score not similarity
                success = True
            except ValueError:
                # get an error with small images that the torchmetrics package seems to advise the wrong larger than size for
                reduced_kernel = True
                mssim_kernel_size -= 2
                _score = 0
            except AssertionError:
                # get an error with small images that the pytorch-ssim package seems to advise the wrong larger than size for
                reduced_kernel = True
                mssim_kernel_size -= 2
                _score = 0
            except RuntimeError:
                run_error = True
                success = True
                _score = 0
            # _metric.reset()
        if reduced_kernel == True:
            print(
                f'NOTE: Reduced MS_SSIM kernel size to {mssim_kernel_size} to deal with image size {im_ref.shape}')
        if run_error == True:
            print(
                f'WARNING: Image size {im_ref.shape} too small to use with MS_SSIM, returning 0')
        return _score

    def _make_metric(self, **kwargs):
        # set up metric
        with warnings.catch_warnings():    # we don't care about the warnings these give
            warnings.simplefilter("ignore")
            if self.return_image:
                _metric = self.metric(**kwargs,
                                      return_full_image=True,
                                      reduction=None)
            else:
                _metric = self.metric(**kwargs)
            _metric.to(self.device)
        return _metric


class NLPD:
    '''Normalised Laplacian pyramid
    Proposed by Valero Laparra et al. https://www.uv.es/lapeva/papers/2016_HVEI.pdf .
    NLPD is an image quality metric based on the transformations associated with the 
    early visual system: local luminance subtraction and local gain control.

    '''

    def __init__(self):
        self.initialised = False   # initialse fully on first __call__ to save load up time
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu")
        self.preproccess_function = _numpy_to_torch_image
        self.nlpd_k = 1

    def __call__(self, im_ref, im_comp, nlpd_k=1, **kwargs):
        '''When an instance is called

        Args:
            im_ref (np.array): Reference image
            im_comp (np.array): Comparison image
            **kwargs: Arbitrary keyword arguments

        Returns:
            score (np.array): NLPD score
        '''
        # see if k has changed
        if self.nlpd_k != nlpd_k:
            self.nlpd_k = nlpd_k
            self.initialised = False
        # load model on first time called
        if self.initialised == False:
            with warnings.catch_warnings():    # we don't care about the warnings these give
                warnings.simplefilter("ignore")
                self.metric = LaplacianPyramid(self.nlpd_k)
            self.metric.to(self.device)
            self.initialised = True

        _check_shapes(im_ref, im_comp)
        im_ref = self.preproccess_function(im_ref).to(
            device=self.device, dtype=torch.float)
        im_comp = self.preproccess_function(im_comp).to(
            device=self.device, dtype=torch.float)
        _score = self.metric(im_ref, im_comp)
        score = _score.cpu().detach().numpy()
        return score


def _make_kernel_odd(value):
    value = int(value)
    if value % 2 == 0:
        value -= 1
    if value <= 0:
        value = 1
    return value



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
import numpy as np
import IQM_Vis


def run():
    # metrics functions must return a single value
    metric = {'MAE': IQM_Vis.metrics.MAE(),
              'MSE': IQM_Vis.metrics.MSE(),
              '1-SSIM': IQM_Vis.metrics.SSIM()
              }

    # metrics images return a numpy image
    metric_images = {
        # 'MSE': IQM_Vis.metrics.MSE(return_image=True),
        # 'SSIM': IQM_Vis.metrics.SSIM(return_image=True)
                     }

    # make dataset list of images
    file_path = os.path.dirname(os.path.abspath(__file__))
    dataset = [IQM_Vis.examples.images.IMAGE1,
               IQM_Vis.examples.images.IMAGE2,
               IQM_Vis.examples.images.IMAGE3]
    
    data = IQM_Vis.dataset_holder(dataset,
                                  metric,
                                  metric_images,
                                  IQM_Vis.utils.load_image,
                                  human_exp_csv=os.path.join(file_path, 'images', 'HIQM.csv'),
                                  )

    # define the transformations
    transformations = {
               'rotation':{'min':-180, 'max':180, 'function':IQM_Vis.transforms.rotation},    # normal input
               'blur':{'min':1, 'max':41, 'normalise':'odd', 'function':IQM_Vis.transforms.blur},  # only odd ints
               'brightness':{'min':-1.0, 'max':1.0, 'function':IQM_Vis.transforms.brightness},   # normal but with float
               # 'threshold':{'min':-40, 'max':40, 'function':IQM_Vis.transforms.binary_threshold},
               'jpg comp.':{'init_value':100, 'min':1, 'max':100, 'function':IQM_Vis.transforms.jpeg_compression},
               }
    # define any parameters that the metrics need (names shared across both metrics and metric_images)
    ssim_params = {'sigma': {'min':0.25, 'max':5.25, 'init_value': 1.5},  # for the guassian kernel
                   # 'kernel_size': {'min':1, 'max':41, 'normalise':'odd', 'init_value': 11},  # ignored if guassian kernel used
                   'k1': {'min':0.01, 'max':0.21, 'init_value': 0.01},
                   'k2': {'min':0.01, 'max':0.21, 'init_value': 0.03}}

    # use the API to create the UI
    IQM_Vis.make_UI(data,
                    transformations,
                    metric_params=ssim_params,
                    )

if __name__ == '__main__':
    run()



''' run all avaiable metrics/ transforms '''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import IQM_Vis

def run():
    # data = IQM_Vis.dataset_holder(IQM_Vis.examples.images.DEFAULT_IMAGES,
    #                               IQM_Vis.metrics.get_all_metrics(),
    #                               IQM_Vis.metrics.get_all_metric_images()
    #                               )


    # IQM_Vis.make_UI(data,
    #                 IQM_Vis.transforms.get_all_transforms(),
    #                 restrict_options=3
    #                 )
    IQM_Vis.make_UI()

if __name__ == '__main__':
    run()



"""
The :mod:`expert.utils.conv` module holds util functions for convolutions, like
padding to maintain the size of the image.
"""
# Author: Alex Hepburn <alex.hepburn@bristol.ac.uk>
# License: new BSD

from typing import List

import math

import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ['pad']


def pad(im_size, filt_size, stride):
    """
    Returns the amount of padding needed on [height, width] to maintain image
    size.

    This function calculates the amount of padding needed to keep the output
    image shape the same as the input image shape.

    Parameters
    ----------
    im_size : List[int]
        List of [height, width] of the image to pad.
    filt_size : int
        The width of the filter being used in the convolution, assumed to be
        square.
    stride : int
        Amount of stride in the convolution.

    Returns
    -------
    padding : List[int]
        Amount of padding needed for []
    """
    out_height = math.ceil(float(im_size[0]) / float(stride))
    out_width  = math.ceil(float(im_size[1]) / float(stride))

    pad_h = max((out_height - 1) * stride + filt_size - im_size[0], 0)
    pad_w = max((out_width - 1) * stride + filt_size - im_size[1], 0)

    pad_top = pad_h // 2
    pad_bottom = pad_h - pad_top
    pad_left = pad_w // 2
    pad_right = pad_w- pad_left

    # Append lists to each other for use in func:`torch.nn.functional.pad`.
    return [pad_left, pad_right, pad_top, pad_bottom]



"""
The :mod:`expert.layers.wavelet_transformation` module holds classes of
layers for wavelet transformations.
"""
# Author: Alex Hepburn <alex.hepburn@bristol.ac.uk>
# License: new BSD

# import expert.utils.fourier as fourier ## needed for SteerableWavelet - includes extra dependancies so wont use for now
from IQM_Vis.metrics.NLPD_torch.utils import conv as conv_utils
# from IQM_Vis.metrics.NLPD_torch.utils import pyramid_filters as pyr_filts
# from IQM_Vis.metrics.NLPD_torch.utils import fourier
# from IQM_Vis.metrics.NLPD_torch.layers import divisive_normalisation as expert_divisive_normalisation
import torch.nn.functional as F
import torch.nn as nn
import torch
# from typing import List, Union

import math

import numpy as np
np.set_printoptions(precision=2)  # TODO: remove


__all__ = ['SteerableWavelet', 'SteerablePyramid', 'LaplacianPyramid',
           'LaplacianPyramidGDN']

LAPLACIAN_FILTER = np.array([[0.0025, 0.0125, 0.0200, 0.0125, 0.0025],
                             [0.0125, 0.0625, 0.1000, 0.0625, 0.0125],
                             [0.0200, 0.1000, 0.1600, 0.1000, 0.0200],
                             [0.0125, 0.0625, 0.1000, 0.0625, 0.0125],
                             [0.0025, 0.0125, 0.0200, 0.0125, 0.0025]],
                            dtype=np.float32)


class LaplacianPyramid(nn.Module):
    def __init__(self, k, dims=3, filt=None, trainable=False):
        super(LaplacianPyramid, self).__init__()
        if filt is None:
            filt = np.reshape(np.tile(LAPLACIAN_FILTER, (dims, 1, 1)),
                              (dims, 1, 5, 5))
        self.k = k
        self.trainable = trainable
        self.dims = dims
        self.filt = nn.Parameter(torch.Tensor(filt), requires_grad=False)
        self.dn_filts, self.sigmas = self.DN_filters()

    def DN_filters(self):
        sigmas = [0.0248, 0.0185, 0.0179, 0.0191, 0.0220, 0.2782]
        dn_filts = []
        dn_filts.append(torch.Tensor(np.reshape([[0, 0.1011, 0],
                                                 [0.1493, 0, 0.1460],
                                                 [0, 0.1015, 0.]]*self.dims,
                                                (self.dims,  1, 3, 3)).astype(np.float32)))

        dn_filts.append(torch.Tensor(np.reshape([[0, 0.0757, 0],
                                                 [0.1986, 0, 0.1846],
                                                 [0, 0.0837, 0]]*self.dims,
                                                (self.dims, 1, 3, 3)).astype(np.float32)))

        dn_filts.append(torch.Tensor(np.reshape([[0, 0.0477, 0],
                                                 [0.2138, 0, 0.2243],
                                                 [0, 0.0467, 0]]*self.dims,
                                                (self.dims, 1, 3, 3)).astype(np.float32)))

        dn_filts.append(torch.Tensor(np.reshape([[0, 0, 0],
                                                 [0.2503, 0, 0.2616],
                                                 [0, 0, 0]]*self.dims,
                                                (self.dims, 1, 3, 3)).astype(np.float32)))

        dn_filts.append(torch.Tensor(np.reshape([[0, 0, 0],
                                                 [0.2598, 0, 0.2552],
                                                 [0, 0, 0]]*self.dims,
                                                (self.dims, 1, 3, 3)).astype(np.float32)))

        dn_filts.append(torch.Tensor(np.reshape([[0, 0, 0],
                                                 [0.2215, 0, 0.0717],
                                                 [0, 0, 0]]*self.dims,
                                                (self.dims, 1, 3, 3)).astype(np.float32)))
        dn_filts = nn.ParameterList([nn.Parameter(x, requires_grad=self.trainable)
                                     for x in dn_filts])
        sigmas = nn.ParameterList([nn.Parameter(torch.Tensor(np.array(x)),
                                                requires_grad=self.trainable) for x in sigmas])
        return dn_filts, sigmas

    def pyramid(self, im):
        out = []
        J = im
        pyr = []
        for i in range(0, self.k):
            J_padding_amount = conv_utils.pad([J.size(2), J.size(3)],
                                              self.filt.size(3), stride=2)
            I = F.conv2d(F.pad(J, J_padding_amount, mode='reflect'), self.filt,
                         stride=2, padding=0, groups=self.dims)
            I_up = F.interpolate(I, size=[J.size(2), J.size(3)],
                                 align_corners=True, mode='bilinear')
            I_padding_amount = conv_utils.pad([I_up.size(2), I_up.size(3)],
                                              self.filt.size(3), stride=1)
            I_up_conv = F.conv2d(F.pad(I_up, I_padding_amount, mode='reflect'),
                                 self.filt, stride=1, padding=0,
                                 groups=self.dims)
            out = J - I_up_conv
            out_padding_amount = conv_utils.pad(
                [out.size(2), out.size(3)], self.dn_filts[i].size(2), stride=1)
            out_conv = F.conv2d(
                F.pad(torch.abs(out), out_padding_amount, mode='reflect'),
                self.dn_filts[i],
                stride=1,
                groups=self.dims)
            out_norm = out / (self.sigmas[i]+out_conv)
            pyr.append(out_norm)
            J = I
        return pyr

    def forward(self, x1, x2):
        # make inputs 3 channel if greyscale
        if x1.shape[1] == 1:
            x1 = torch.cat([x1, x1, x1], dim=1)
        if x2.shape[1] == 1:
            x2 = torch.cat([x2, x2, x2], dim=1)
        y1 = self.pyramid(x1)
        y2 = self.pyramid(x2)
        total = []
        # Calculate difference in perceptual space (Tensors are stored
        # strangley to avoid needing to pad tensors)
        for z1, z2 in zip(y1, y2):
            diff = (z1 - z2) ** 2
            sqrt = torch.sqrt(torch.mean(diff, (1, 2, 3)))
            total.append(sqrt)
        return torch.norm(torch.stack(total), 0.6)


# not currently used to commenting for coverage

# class LaplacianPyramidGDN(nn.Module):
#     def __init__(self, k, dims=3, filt=None):
#         super(LaplacianPyramidGDN, self).__init__()
#         if filt is None:
#             filt = np.tile(LAPLACIAN_FILTER, (dims, 1, 1))
#             filt = np.reshape(np.tile(LAPLACIAN_FILTER, (dims, 1, 1)),
#                               (dims, 1, 5, 5))
#         self.k = k
#         self.dims = dims
#         self.filt = nn.Parameter(torch.Tensor(filt))
#         self.filt.requires_grad = False
#         self.gdns = nn.ModuleList([expert_divisive_normalisation.GDN(
#             dims, apply_independently=True) for i in range(self.k)])
#         self.pad_one = nn.ReflectionPad2d(1)
#         self.pad_two = nn.ReflectionPad2d(2)
#         self.mse = nn.MSELoss(reduction='none')
#         self.upsample = nn.Upsample(scale_factor=2, mode='bilinear',
#                                     align_corners=True)

#     def pyramid(self, im):
#         J = im
#         pyr = []
#         for i in range(0, self.k):
#             I = F.conv2d(self.pad_two(J), self.filt, stride=2, padding=0,
#                          groups=self.dims)
#             I_up = self.upsample(I)
#             I_up_conv = F.conv2d(self.pad_two(I_up), self.filt, stride=1,
#                                  padding=0, groups=self.dims)
#             if J.size() != I_up_conv.size():
#                 I_up_conv = torch.nn.functional.interpolate(
#                     I_up_conv, [J.size(2), J.size(3)])
#             pyr.append(self.gdns[i](J - I_up_conv))
#             J = I
#         return pyr

#     def compare(self, x1, x2):
#         y1 = self.pyramid(x1)
#         y2 = self.pyramid(x2)
#         total = []
#         # Calculate difference in perceptual space (Tensors are stored
#         # strangley to avoid needing to pad tensors)
#         for z1, z2 in zip(y1, y2):
#             diff = (z1 - z2) ** 2
#             sqrt = torch.sqrt(torch.mean(diff, (1, 2, 3)))
#             total.append(sqrt)
#         return torch.norm(torch.stack(total), 0.6)


# class SteerableWavelet(nn.Module):
#     """
#     Steerable wavelet pyramid.

#     Performs a high-pass and low-pass filtering. The low-pass filtered image
#     is then passed through ``wavelets`` wavelet transformations, each
#     with a different orientation. This is done at ``scales`` dfferent scales
#     where each scale is downsampled by ``downsampled``.

#     TODO : pass dimension parameter into __init__ and precompute log_rad
#            and angle.

#     Parameters
#     ----------
#     stages : int, optional (default=4)
#         Number of stages to be used in the pyramid.
#     order: int, optional (default=3)
#         Defined angular functions. Angular functions are
#         `cos(theta-order\pi/(order+1))^order` (order is one less than the
#         number of orientation bands).
#     twidth : int, optional(default=1)
#         width of the transition region of the radial lowpass function in
#         octaves.

#     Raises
#     ------

#     Attributes
#     ----------
#     num_orientations : integer
#     harmincs : torch.Tensor
#     angles : torch.Tensor
#     steer_matrix : torch.Tensor
#     Xrcos : torch.Tensor
#     Yrcos : torch.Tensor
#     YIrcos : torch.Tensor
#     Xcosn : torch.Tensor
#     const : float
#     Ycosn : torch.Tensor
#     """

#     def __init__(self,
#                  stages: int = 4,
#                  order: int = 3,
#                  twidth: int = 1):
#         """
#         Constructs a ``SteerableWavelet`` class.
#         """
#         super(SteerableWavelet, self).__init__()
#         assert self._validate_input(stages, order, twidth)
#         self.stages = stages
#         self.num_orientations = order + 1
#         self.twidth = twidth

#         if self.num_orientations % 2 == 0:
#             self.harmonics = torch.arange(0, self.num_orientations/2)*2+1
#         else:
#             self.harmonics = torch.arange(0, order)*2

#         self.angles = torch.arange(0, order+1) * \
#             math.pi / self.num_orientations
#         self.steer_matrix = fourier.steer_to_harmonics(
#             self.harmonics, self.angles, 'sin')

#         # CONSTANTS
#         self.Xrcos, self.Yrcos = fourier.raised_cosine(
#             twidth, -twidth/2, func_min=0.0, func_max=1.0, size=10)
#         self.Yrcos = torch.sqrt(self.Yrcos)
#         self.YIrcos = torch.sqrt(1.0 - self.Yrcos**2)
#         self.Xcosn = math.pi * torch.arange(-2047, 1024+2) / 1024  # (-2*pi:pi]
#         self.const = 2**(2*order) * math.factorial(order)**2 \
#             / (self.num_orientations * math.factorial(order*2))
#         self.Ycosn = math.sqrt(self.const) * torch.cos(self.Xcosn)**order

#         # Orientation angles
#         self.band_angles = torch.Tensor(
#             [self.Xcosn[0] + math.pi*i/self.num_orientations
#              for i in range(self.num_orientations)])

#     def _validate_input(self,
#                         stages: int,
#                         order: int,
#                         twidth: int) -> bool:
#         """
#         Validates input of the steerable wavelet class.

#         For the description of the input parameters and exceptions raised by
#         this function, please see the documentation of the
#         :class:`expert.models.pyramids.SteerableWavelet` class.

#         Returns
#         -------
#         is_valid
#             ``True`` if input is valid, ``False`` otherwise.
#         """
#         is_valid = False

#         if not isinstance(stages, int) or stages <= 0:
#             raise TypeError('stages parameter must be an integer greater than '
#                             '0.')

#         if not isinstance(order, int) or order <= 0:
#             raise TypeError('order parameter must be an integer '
#                             'greater than 0.')

#         if not isinstance(twidth, int) or twidth <= 0:
#             raise TypeError('twidth parameter must be an integer greater than '
#                             '0.')

#         is_valid = True
#         return is_valid

#     def meshgrid_angle(self,
#                        dims: List[int]) -> torch.Tensor:
#         """
#         Computes meshgrid of input dimensions and angles

#         Takes the dimensions of an image, and computes a meshgrid over these
#         values. Then computes the angles and log radians of this meshgrid.

#         Parameters
#         ----------
#         dims : List[Integer]
#             List of length two sepcificy [h, w] of image.

#         Returns
#         -------
#         angle : torch.Tensor
#             Angle of meshgrid.
#         log_rad : torch.Tensor
#             Angle in log radians.
#         """
#         ctr = [math.ceil((d+0.5)/2) for d in dims]
#         vectors = [
#             (torch.arange(1, dims[i]+1)-ctr[i]) / (dims[i]/2) for i in range(2)]
#         yramp, xramp = torch.meshgrid(vectors[1], vectors[0])
#         angle = torch.atan2(yramp, xramp)
#         log_rad = torch.sqrt(xramp**2 + yramp**2)

#         # Replace middle element in matrix
#         log_rad[ctr[0]-1, ctr[1]-1] = log_rad[ctr[0]-1, ctr[1]-2]
#         log_rad = torch.log2(log_rad)

#         return angle, log_rad

#     def _check_height(self,
#                       size_x: torch.Size) -> bool:
#         """
#         Checks if dimensions of image is compatible with height of pyramid.

#         Calculates the max height a pyramid with the given `size_x` and
#         compares this with the height intialised with this pyramid.

#         Parameters
#         ----------
#         size_x : torch.Size
#             Size of input x. First dimension is batch dimension.

#         Raises
#         ------
#         ValueError:
#             If the maximum height of pyramid for `size_x` is larger than the
#             height of `self.stages`

#         Returns
#         -------
#         is_valid : boolean
#             ``True`` the `size_x` dimensions are compatible with `self.stages`,
#             else ``False``.
#         """
#         is_valid = False
#         dims = list(size_x)[1:]
#         max_height = math.floor(math.log2(min(dims))) - 2
#         if self.stages > max_height:
#             raise ValueError('Input maximum number of stages is %d but number '
#                              'of pyramid stages is %d. Please use larger input '
#                              'images or initialise pyramid with different '
#                              'number of stages.' % (max_height, self.stages))

#         is_valid = True
#         return is_valid

#     def _complex_number_product(self,
#                                 x: torch.Tensor) -> torch.Tensor:
#         """
#         Multiple tensor with imaginary number.

#         If `o` is the order within the pyramid, then output `y` is given by
#         .. math::
#             \mathbf{y} = ((-\sqrt{-1})^o) * \mathbf{x}.

#         The value of (-sqrt(-1))^order cycles through values. List below are
#         the possible cases.
#         If order % 4 == 0, (-sqrt(-1))^order = 1
#         If order % 4 == 1, (-sqrt(-1))^order = -j
#         If order % 4 == 2, (-sqrt(-1))^order = -1
#         If order % 4 == 3, (-sqrt(-1))^order = j

#         Parameters
#         ----------
#         x : torch.Tensor
#             Input tensor.

#         Returns
#         -------
#         y: torch.Tensor
#             Output tensor.
#         """
#         value = ((self.num_orientations-1) % 4)
#         if value == 0:  # x * 1
#             return x
#         else:
#             real, imag = torch.unbind(x, -1)
#             if value == 1:  # x * -j
#                 real_new = imag
#                 imag_new = -real
#             elif value == 2:  # x * -1
#                 real_new = -real
#                 imag_new = -imag
#             elif value == 3:  # x * j
#                 real_new = -imag
#                 imag_new = real
#             return torch.stack((real_new, imag_new), -1)

#     def forward(
#         self,
#         x: torch.tensor,
#         upsample_output: bool = False
#     ) -> Union[List[torch.Tensor], torch.Tensor]:
#         """
#         Forward pass of the pyramid.

#         This function returns a lit of Tensors for the subbands in each stage
#         of the pyramid or, if ``upsample_output`` is ``True``, a Tensor
#         containing every subband at every level that has been upsampled to be
#         the same size as the input Tensor.

#         Parameters
#         ----------
#         x : torch.Tensor
#             Input to the pyramid.
#         upsample_output : boolean
#             If ``True`` then the every subband will be upsampled to be the same
#             size as the input and then stacked to be a singular tensor.

#         Raises
#         ------
#         TypeError:
#             Input parameter ``x`` is not of dtype torch.float.

#         Returns
#         -------
#         pyramid : Union[List[torch.Tensor], Tensor]
#             List of tensors, where the first tensor is the high pass residual,
#             and from thereon each entry contains the subbands at each stage of
#             the pyramid. The low pass residual is the last element in the
#             pyramid. If ``upsample_output`` is ``True`` then this will be one
#             Tensor where each subband has been upsample to be the same
#             dimensions as the input.
#         """
#         assert self._check_height(x.size())

#         dims = [x.size(1), x.size(2)]
#         self.angle, self.log_rad = self.meshgrid_angle(dims=dims)

#         low_mask = fourier.point_operation_filter(
#             self.log_rad, self.YIrcos, origin=self.Xrcos[0],
#             increment=self.Xrcos[1]-self.Xrcos[0]).to(x.device)
#         self.low_mask = low_mask.view(1, low_mask.size(0), low_mask.size(1), 1)

#         high_mask = fourier.point_operation_filter(
#             self.log_rad, self.Yrcos, origin=self.Xrcos[0],
#             increment=self.Xrcos[1]-self.Xrcos[0]).to(x.device)
#         self.high_mask = high_mask.view(
#             1, high_mask.size(0), high_mask.size(1), 1)

#         # Shape [batch, height, width, 2] where last dimension is real and
#         # imaginary part of Fourier transform.
#         discrete_fourier_image = fourier.fftshift(
#             torch.rfft(x, signal_ndim=2, onesided=False), dim=(-3, -2))

#         # Calculate highpass and lowpass of the image.
#         low_pass = discrete_fourier_image * self.low_mask
#         high_pass = discrete_fourier_image * self.high_mask

#         # First element in the pyramid will be the high pass residual
#         high_pass_residual = torch.irfft(fourier.ifftshift(
#             high_pass, dim=(-3, -2)), signal_ndim=2, onesided=False)
#         pyramid = [high_pass_residual]

#         # Intiailised parameters that are going to change each stage
#         Xrcos = self.Xrcos
#         log_rad = self.log_rad
#         angle = self.angle

#         # Stages in pyramid
#         for h in range(0, self.stages):
#             Xrcos = Xrcos - math.log2(2)
#             himask = fourier.point_operation_filter(
#                 log_rad, self.Yrcos, origin=Xrcos[0],
#                 increment=Xrcos[1]-Xrcos[0])
#             himask = himask.view(1, himask.size(0), himask.size(1), 1)

#             # Orientation band masks
#             angle_masks = torch.stack(
#                 [fourier.point_operation_filter(angle, self.Ycosn,
#                                                 origin=a, increment=self.Xcosn[1]-self.Xcosn[0])
#                  for a in self.band_angles])
#             fourier_band = (low_pass.permute(3, 0, 1, 2) *
#                             angle_masks).unsqueeze(0).permute(0, 2, 3, 4, 1)
#             # TODO: computation only works for batch_size=1
#             fourier_band = fourier_band * himask
#             fourier_band = self._complex_number_product(fourier_band)
#             # Collapse from [batch, bands, height, width, 2] to
#             # [batch*bands, height, width, 2] for ifft.
#             bands = torch.irfft(fourier.ifftshift(
#                 fourier_band.view(-1, fourier_band.size(2),
#                                   fourier_band.size(3), 2), dim=(-3, -2)), signal_ndim=2,
#                                 onesided=False)
#             # Expand back out to [batch, bands, height, width] after doing
#             # inverse fast fourier transform.
#             bands = bands.view(x.size(0), self.num_orientations, bands.size(1),
#                                bands.size(2))
#             pyramid.append(bands)

#             # Subsampling operations
#             dims = [low_pass.size(1), low_pass.size(2)]
#             ctr = [math.ceil((d+0.5)/2) for d in dims]
#             lodims = [math.ceil((d-0.5)/2) for d in dims]
#             loctr = [math.ceil((d+0.5)/2) for d in lodims]

#             lostart = [c-l for c, l in zip(ctr, loctr)]
#             loend = [ls+lo for ls, lo in zip(lostart, lodims)]

#             log_rad = log_rad[lostart[0]:loend[0], lostart[1]:loend[1]]
#             angle = angle[lostart[0]:loend[0], lostart[1]:loend[1]]
#             low_pass = low_pass[:, lostart[0]:loend[0], lostart[1]:loend[1]]

#             # Low pass to use in the next stage
#             low_mask = fourier.point_operation_filter(
#                 log_rad, self.YIrcos, origin=Xrcos[0],
#                 increment=Xrcos[1]-Xrcos[0]).to(x.device)
#             low_mask = low_mask.view(1, low_mask.size(0), low_mask.size(1), 1)
#             low_pass = low_pass * low_mask

#         low_pass_residual = torch.irfft(fourier.ifftshift(
#             low_pass, dim=(-3, -2)), signal_ndim=2, onesided=False)
#         pyramid.append(low_pass_residual)

#         if upsample_output:
#             original_size = [high_pass.size(2), high_pass.size(3)]
#             pyr_upsample = [
#                 F.interpolate(stage, size=original_size) for stage in pyramid
#             ]
#             pyramid = torch.cat(pyr_upsample, dim=1)

#         return pyramid


# class SteerablePyramid(nn.Module):
#     """
#     Steerable pyramid model implemented in spatial domain, introduced in
#     [SIMON1995PYR]_.

#     Parameters
#     ----------
#     stages : int, optional (default=4)
#         Number of stages to be used in the pyramid.
#     num_orientations: int, optional (default=2)
#         Number of orientations to be used at each stage of the pyramid (number
#         of subbands). If ``pretrained`` is ``True`` then this must be 2 as the
#         pretarined weights from the original implementation use 2 subbands.
#     pretrained : bool, optional (default=False)
#         Whether to load the pretrained filters, specified in the original paper
#         [SIMON1995PYR]_.

#     Raises
#     ------


#     Attributes
#     ----------

#     .. [SIMON1995PYR] E P Simoncelli and W T Freeman, "The Steerable Pyramid:
#        A Flexible Architecture for Multi-Scale Derivative Computation," Second
#        Int'l Conf on Image Processing, Washington, DC, Oct 1995.

#     TODO: implement an initialisation that sets ``num_orientations`` filters
#           that are rotated a different amounts, and optimise only over the
#           amount of rotation for each filter.
#     """

#     def __init__(self,
#                  stages: int = 4,
#                  num_orientations: int = 2,
#                  pretrained: bool = False) -> None:
#         """
#         Constructs a ``SteerablePyramid`` class.
#         """
#         super(SteerablePyramid, self).__init__()
#         assert self._validate_input(stages, num_orientations, pretrained)
#         self.stages = stages
#         self.num_orientations = num_orientations

#         self.lo0filt = nn.Parameter(torch.ones(1, 1, 9, 9))
#         self.hi0filt = nn.Parameter(torch.ones(1, 1, 9, 9))
#         self.lofilt = nn.Parameter(torch.ones(1, 1, 17, 17))
#         self.bfilts = nn.Parameter(torch.ones(self.num_orientations, 1, 9, 9))

#         for param in self.parameters():
#             torch.nn.init.normal_(param)

#         if pretrained:
#             filters = pyr_filts.STEERABLE_SPATIAL_FILTERS_1
#             with torch.no_grad():
#                 self.lo0filt.data = filters['lo0filt']
#                 self.hi0filt.data = filters['hi0filt']
#                 self.lofilt.data = filters['lofilt']
#                 self.bfilts.data = filters['bfilts']

#     def _validate_input(self,
#                         stages: int,
#                         num_orientations: int,
#                         pretrained: bool) -> bool:
#         """
#         Validates input of the steerable pyramid class.

#         For the description of the input parameters and exceptions raised by
#         this function, please see the documentation of the
#         :class:`expert.models.pyramids.SteerablePyramid` class.

#         Returns
#         -------
#         is_valid
#             ``True`` if input is valid, ``False`` otherwise.
#         """
#         is_valid = False

#         if not isinstance(stages, int) or stages <= 0:
#             raise TypeError('stages parameter must be an integer greater than '
#                             '0.')

#         if not isinstance(num_orientations, int) or num_orientations <= 0:
#             raise TypeError('num_orientations parameter must be an integer '
#                             'greater than 0.')

#         if not isinstance(pretrained, bool):
#             raise TypeError('pretrained parameter must be a boolean.')

#         if pretrained and num_orientations != 2:
#             raise ValueError('To use the pretrained network, num_orientations '
#                              'must be 2.')

#         is_valid = True
#         return is_valid

#     def forward(
#         self,
#         x: torch.tensor,
#         upsample_output: bool = False
#     ) -> Union[List[torch.Tensor], torch.Tensor]:
#         """
#         Forward pass of the pyramid.

#         This function returns a lit of Tensors for the subbands in each stage
#         of the pyramid or, if ``upsample_output`` is ``True``, a Tensor
#         containing every subband at every level that has been upsampled to be
#         the same size as the input Tensor.

#         Parameters
#         ----------
#         x : torch.Tensor
#             Input to the pyramid.
#         upsample_output : boolean
#             If ``True`` then the every subband will be upsampled to be the same
#             size as the input and then stacked to be a singular tensor.

#         Raises
#         ------
#         TypeError:
#             Input parameter ``x`` is not of dtype torch.float.

#         Returns
#         -------
#         pyramid : Union[List[torch.Tensor], Tensor]
#             List of tensors, where the first tensor is the high pass residual,
#             and from thereon each entry contains the subbands at each stage of
#             the pyramid. The low pass residual is the last element in the
#             pyramid. If ``upsample_output`` is ``True`` then this will be one
#             Tensor where each subband has been upsample to be the same
#             dimensions as the input.
#         """
#         if x.dtype != torch.float32:
#             raise TypeError('Input x must be of type torch.float32.')

#         padded_x = F.pad(x,
#                          pad=conv_utils.pad([x.size(2), x.size(3)], 9, 1),
#                          mode='reflect')
#         low_pass = F.conv2d(padded_x, self.lo0filt)
#         high_pass = F.conv2d(padded_x, self.hi0filt)

#         pyramid = []
#         pyramid.append(high_pass)

#         for h in range(0, self.stages):
#             image_size = [low_pass.size(2), low_pass.size(3)]
#             padded_lowpass = F.pad(low_pass,
#                                    pad=conv_utils.pad(image_size, 9, 1),
#                                    mode='reflect')
#             subbands = F.conv2d(padded_lowpass, self.bfilts, groups=1)
#             padded_lowpass = F.pad(low_pass,
#                                    pad=conv_utils.pad(image_size, 17, 1),
#                                    mode='reflect')
#             low_pass = F.conv2d(padded_lowpass, self.lofilt, stride=[2, 2])
#             pyramid.append(subbands)
#         pyramid.append(low_pass)

#         if upsample_output:
#             original_size = [high_pass.size(2), high_pass.size(3)]
#             pyr_upsample = [
#                 F.interpolate(stage, size=original_size) for stage in pyramid
#             ]
#             pyramid = torch.cat(pyr_upsample, dim=1)

#         return pyramid, high_pass



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from PyQt6.QtWidgets import QWidget


def add_layout_to_tab(tab, layout, name):
    _tab = QWidget()   # QTabWidget only accepts widgets not layouts so need to use this as a workaround
    _tab.setLayout(layout)
    tab.addTab(_tab, name)



import os
import glob

dir = os.path.dirname(os.path.abspath(__file__))

KODAK_IMAGES = glob.glob(os.path.join(dir, '*'))
# remove and folders
KODAK_IMAGES = [f for f in KODAK_IMAGES if os.path.isfile(f)]
KODAK_IMAGES.sort()



# taken from https://github.com/VainF/pytorch-msssim/blob/master/pytorch_msssim/ssim.py
# Copyright 2020 by Gongfan Fang, Zhejiang University.
# All rights reserved.
# Author: Gongfan Fang
# Licence: MIT

import warnings
from typing import List, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from torch import Tensor


def _fspecial_gauss_1d(size: int, sigma: float) -> Tensor:
    r"""Create 1-D gauss kernel
    Args:
        size (int): the size of gauss kernel
        sigma (float): sigma of normal distribution
    Returns:
        torch.Tensor: 1D kernel (1 x 1 x size)
    """
    coords = torch.arange(size, dtype=torch.float)
    coords -= size // 2

    g = torch.exp(-(coords ** 2) / (2 * sigma ** 2))
    g /= g.sum()

    return g.unsqueeze(0).unsqueeze(0)


def gaussian_filter(input: Tensor, win: Tensor) -> Tensor:
    r""" Blur input with 1-D kernel
    Args:
        input (torch.Tensor): a batch of tensors to be blurred
        window (torch.Tensor): 1-D gauss kernel
    Returns:
        torch.Tensor: blurred tensors
    """
    assert all([ws == 1 for ws in win.shape[1:-1]]), win.shape
    if len(input.shape) == 4:
        conv = F.conv2d
    elif len(input.shape) == 5:
        conv = F.conv3d
    else:
        raise NotImplementedError(input.shape)

    C = input.shape[1]
    out = input
    for i, s in enumerate(input.shape[2:]):
        if s >= win.shape[-1]:
            out = conv(out, weight=win.transpose(2 + i, -1), stride=1, padding=0, groups=C)
        else:
            warnings.warn(
                f"Skipping Gaussian Smoothing at dimension 2+{i} for input: {input.shape} and win size: {win.shape[-1]}"
            )

    return out


def _ssim(
    X: Tensor,
    Y: Tensor,
    data_range: float,
    win: Tensor,
    size_average: bool = True,
    K: Union[Tuple[float, float], List[float]] = (0.01, 0.03)
) -> Tuple[Tensor, Tensor]:
    r""" Calculate ssim index for X and Y

    Args:
        X (torch.Tensor): images
        Y (torch.Tensor): images
        data_range (float or int): value range of input images. (usually 1.0 or 255)
        win (torch.Tensor): 1-D gauss kernel
        size_average (bool, optional): if size_average=True, ssim of all images will be averaged as a scalar

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: ssim results.
    """
    K1, K2 = K
    # batch, channel, [depth,] height, width = X.shape
    compensation = 1.0

    C1 = (K1 * data_range) ** 2
    C2 = (K2 * data_range) ** 2

    win = win.to(X.device, dtype=X.dtype)

    mu1 = gaussian_filter(X, win)
    mu2 = gaussian_filter(Y, win)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2

    sigma1_sq = compensation * (gaussian_filter(X * X, win) - mu1_sq)
    sigma2_sq = compensation * (gaussian_filter(Y * Y, win) - mu2_sq)
    sigma12 = compensation * (gaussian_filter(X * Y, win) - mu1_mu2)

    cs_map = (2 * sigma12 + C2) / (sigma1_sq + sigma2_sq + C2)  # set alpha=beta=gamma=1
    ssim_map = ((2 * mu1_mu2 + C1) / (mu1_sq + mu2_sq + C1)) * cs_map

    ssim_per_channel = torch.flatten(ssim_map, 2).mean(-1)
    cs = torch.flatten(cs_map, 2).mean(-1)
    return ssim_per_channel, cs


# not currently used to commenting for coverage

# def ssim(
#     X: Tensor,
#     Y: Tensor,
#     data_range: float = 255,
#     size_average: bool = True,
#     win_size: int = 11,
#     win_sigma: float = 1.5,
#     win: Optional[Tensor] = None,
#     K: Union[Tuple[float, float], List[float]] = (0.01, 0.03),
#     nonnegative_ssim: bool = False,
# ) -> Tensor:
#     r""" interface of ssim
#     Args:
#         X (torch.Tensor): a batch of images, (N,C,H,W)
#         Y (torch.Tensor): a batch of images, (N,C,H,W)
#         data_range (float or int, optional): value range of input images. (usually 1.0 or 255)
#         size_average (bool, optional): if size_average=True, ssim of all images will be averaged as a scalar
#         win_size: (int, optional): the size of gauss kernel
#         win_sigma: (float, optional): sigma of normal distribution
#         win (torch.Tensor, optional): 1-D gauss kernel. if None, a new kernel will be created according to win_size and win_sigma
#         K (list or tuple, optional): scalar constants (K1, K2). Try a larger K2 constant (e.g. 0.4) if you get a negative or NaN results.
#         nonnegative_ssim (bool, optional): force the ssim response to be nonnegative with relu

#     Returns:
#         torch.Tensor: ssim results
#     """
#     if not X.shape == Y.shape:
#         raise ValueError(f"Input images should have the same dimensions, but got {X.shape} and {Y.shape}.")

#     for d in range(len(X.shape) - 1, 1, -1):
#         X = X.squeeze(dim=d)
#         Y = Y.squeeze(dim=d)

#     if len(X.shape) not in (4, 5):
#         raise ValueError(f"Input images should be 4-d or 5-d tensors, but got {X.shape}")

#     #if not X.type() == Y.type():
#     #    raise ValueError(f"Input images should have the same dtype, but got {X.type()} and {Y.type()}.")

#     if win is not None:  # set win_size
#         win_size = win.shape[-1]

#     if not (win_size % 2 == 1):
#         raise ValueError("Window size should be odd.")

#     if win is None:
#         win = _fspecial_gauss_1d(win_size, win_sigma)
#         win = win.repeat([X.shape[1]] + [1] * (len(X.shape) - 1))

#     ssim_per_channel, cs = _ssim(X, Y, data_range=data_range, win=win, size_average=False, K=K)
#     if nonnegative_ssim:
#         ssim_per_channel = torch.relu(ssim_per_channel)
    
#     if size_average:
#         return ssim_per_channel.mean()
#     else:
#         return ssim_per_channel.mean(1)


def ms_ssim(
    X: Tensor,
    Y: Tensor,
    data_range: float = 255,
    size_average: bool = True,
    win_size: int = 11,
    win_sigma: float = 1.5,
    win: Optional[Tensor] = None,
    weights: Optional[List[float]] = None,
    K: Union[Tuple[float, float], List[float]] = (0.01, 0.03)
) -> Tensor:
    r""" interface of ms-ssim
    Args:
        X (torch.Tensor): a batch of images, (N,C,[T,]H,W)
        Y (torch.Tensor): a batch of images, (N,C,[T,]H,W)
        data_range (float or int, optional): value range of input images. (usually 1.0 or 255)
        size_average (bool, optional): if size_average=True, ssim of all images will be averaged as a scalar
        win_size: (int, optional): the size of gauss kernel
        win_sigma: (float, optional): sigma of normal distribution
        win (torch.Tensor, optional): 1-D gauss kernel. if None, a new kernel will be created according to win_size and win_sigma
        weights (list, optional): weights for different levels
        K (list or tuple, optional): scalar constants (K1, K2). Try a larger K2 constant (e.g. 0.4) if you get a negative or NaN results.
    Returns:
        torch.Tensor: ms-ssim results
    """
    if not X.shape == Y.shape:
        raise ValueError(f"Input images should have the same dimensions, but got {X.shape} and {Y.shape}.")

    for d in range(len(X.shape) - 1, 1, -1):
        X = X.squeeze(dim=d)
        Y = Y.squeeze(dim=d)

    #if not X.type() == Y.type():
    #    raise ValueError(f"Input images should have the same dtype, but got {X.type()} and {Y.type()}.")

    if len(X.shape) == 4:
        avg_pool = F.avg_pool2d
    elif len(X.shape) == 5:
        avg_pool = F.avg_pool3d
    else:
        raise ValueError(f"Input images should be 4-d or 5-d tensors, but got {X.shape}")

    if win is not None:  # set win_size
        win_size = win.shape[-1]

    if not (win_size % 2 == 1):
        raise ValueError("Window size should be odd.")

    smaller_side = min(X.shape[-2:])
    assert smaller_side > (win_size - 1) * (
        2 ** 4
    ), "Image size should be larger than %d due to the 4 downsamplings in ms-ssim" % ((win_size - 1) * (2 ** 4))

    if weights is None:
        weights = [0.0448, 0.2856, 0.3001, 0.2363, 0.1333]
    weights_tensor = X.new_tensor(weights)

    if win is None:
        win = _fspecial_gauss_1d(win_size, win_sigma)
        win = win.repeat([X.shape[1]] + [1] * (len(X.shape) - 1))

    levels = weights_tensor.shape[0]
    mcs = []
    for i in range(levels):
        ssim_per_channel, cs = _ssim(X, Y, win=win, data_range=data_range, size_average=False, K=K)

        if i < levels - 1:
            mcs.append(torch.relu(cs))
            padding = [s % 2 for s in X.shape[2:]]
            X = avg_pool(X, kernel_size=2, padding=padding)
            Y = avg_pool(Y, kernel_size=2, padding=padding)

    ssim_per_channel = torch.relu(ssim_per_channel)  # type: ignore  # (batch, channel)
    mcs_and_ssim = torch.stack(mcs + [ssim_per_channel], dim=0)  # (level, batch, channel)
    ms_ssim_val = torch.prod(mcs_and_ssim ** weights_tensor.view(-1, 1, 1), dim=0)

    if size_average:
        return ms_ssim_val.mean()
    else:
        return ms_ssim_val.mean(1)


# not currently used to commenting for coverage

# class SSIM(torch.nn.Module):
#     def __init__(
#         self,
#         data_range: float = 255,
#         size_average: bool = True,
#         win_size: int = 11,
#         win_sigma: float = 1.5,
#         channel: int = 3,
#         spatial_dims: int = 2,
#         K: Union[Tuple[float, float], List[float]] = (0.01, 0.03),
#         nonnegative_ssim: bool = False,
#     ) -> None:
#         r""" class for ssim
#         Args:
#             data_range (float or int, optional): value range of input images. (usually 1.0 or 255)
#             size_average (bool, optional): if size_average=True, ssim of all images will be averaged as a scalar
#             win_size: (int, optional): the size of gauss kernel
#             win_sigma: (float, optional): sigma of normal distribution
#             channel (int, optional): input channels (default: 3)
#             K (list or tuple, optional): scalar constants (K1, K2). Try a larger K2 constant (e.g. 0.4) if you get a negative or NaN results.
#             nonnegative_ssim (bool, optional): force the ssim response to be nonnegative with relu.
#         """

#         super(SSIM, self).__init__()
#         self.win_size = win_size
#         self.win = _fspecial_gauss_1d(win_size, win_sigma).repeat([channel, 1] + [1] * spatial_dims)
#         self.size_average = size_average
#         self.data_range = data_range
#         self.K = K
#         self.nonnegative_ssim = nonnegative_ssim

#     def forward(self, X: Tensor, Y: Tensor) -> Tensor:
#         return ssim(
#             X,
#             Y,
#             data_range=self.data_range,
#             size_average=self.size_average,
#             win=self.win,
#             K=self.K,
#             nonnegative_ssim=self.nonnegative_ssim,
#         )


# class MS_SSIM(torch.nn.Module):
#     def __init__(
#         self,
#         data_range: float = 255,
#         size_average: bool = True,
#         win_size: int = 11,
#         win_sigma: float = 1.5,
#         channel: int = 3,
#         spatial_dims: int = 2,
#         weights: Optional[List[float]] = None,
#         K: Union[Tuple[float, float], List[float]] = (0.01, 0.03),
#     ) -> None:
#         r""" class for ms-ssim
#         Args:
#             data_range (float or int, optional): value range of input images. (usually 1.0 or 255)
#             size_average (bool, optional): if size_average=True, ssim of all images will be averaged as a scalar
#             win_size: (int, optional): the size of gauss kernel
#             win_sigma: (float, optional): sigma of normal distribution
#             channel (int, optional): input channels (default: 3)
#             weights (list, optional): weights for different levels
#             K (list or tuple, optional): scalar constants (K1, K2). Try a larger K2 constant (e.g. 0.4) if you get a negative or NaN results.
#         """

#         super(MS_SSIM, self).__init__()
#         self.win_size = win_size
#         self.win = _fspecial_gauss_1d(win_size, win_sigma).repeat([channel, 1] + [1] * spatial_dims)
#         self.size_average = size_average
#         self.data_range = data_range
#         self.weights = weights
#         self.K = K

#     def forward(self, X: Tensor, Y: Tensor) -> Tensor:
#         return ms_ssim(
#             X,
#             Y,
#             data_range=self.data_range,
#             size_average=self.size_average,
#             win=self.win,
#             weights=self.weights,
#             K=self.K,
#         )


# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

'''
Test that all of the metrics work as expected
'''
import os
import numpy as np
from IQM_Vis.metrics import get_all_metrics, get_all_metric_images
from IQM_Vis.utils import load_image
from IQM_Vis.examples.images import IMAGE2


def _get_test_image():
    return load_image(IMAGE2)


def test_all_metrics_initialise():
    # get all metrics 
    all_metrics = get_all_metrics()
    for name, metric in all_metrics.items():
        # assert that the metric is callable
        assert callable(metric)


def test_all_metrics_images_initialise():
    # get all metrics 
    all_metrics = get_all_metric_images()
    for name, metric in all_metrics.items():
        # assert that the metric is callable
        assert callable(metric)


def test_all_metrics_process_image():
    # make a mock image
    image = _get_test_image()
    # test that the default value doesn't change the image
    all_metrics = get_all_metrics()
    for name, metric in all_metrics.items():
        val = metric(image, image)
        if isinstance(val, np.ndarray):
            assert val.size == 1
        else:
            assert isinstance(val, float) or isinstance(val, int)


def test_all_metric_images_process_image():
    # make a mock image
    image = _get_test_image()
    # test that the default value doesn't change the image
    all_metrics = get_all_metric_images()
    for name, metric in all_metrics.items():
        img = metric(image, image)
        assert len(img.shape) == 3 or len(img.shape) == 2


if __name__ == '__main__':
    test_all_metrics_initialise()
    test_all_metrics_images_initialise()
    test_all_metrics_process_image()



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from setuptools import setup, find_packages
import os


def get_version():
    info = {}
    with open(os.path.join('IQM_Vis', 'version.py')) as ver_file:
        exec(ver_file.read(), info)
    return info['__version__']


def get_long_description():
    with open('README.md', encoding="utf-8") as f:
        text = f.read()
    return text


def dependencies_from_file(file_path):
    required = []
    with open(file_path) as f:
        for l in f.readlines():
            l_c = l.strip()
            # get not empty lines and ones that do not start with python
            # comment "#" (preceded by any number of white spaces)
            if l_c and not l_c.startswith('#'):
                required.append(l_c)
    return required


setup(name='IQM-Vis',
      version=get_version(),
      packages=find_packages(),
      include_package_data=True,
      install_requires=dependencies_from_file('./requirements.txt'),
      author="Matt Clifford",
      author_email="matt.clifford@bristol.ac.uk",
      description="Extendable user interface for the assessment of transformations on image metrics.",
      long_description=get_long_description(),
      long_description_content_type='text/markdown',
      )



'''
UI image functions
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
from functools import partial
import imghdr
import glob
import numpy as np
import pandas as pd
from PyQt6.QtWidgets import QApplication, QFileDialog
from PyQt6.QtCore import pyqtSignal, QThread
import IQM_Vis
from IQM_Vis.utils import gui_utils, plot_utils, image_utils, save_utils

# sub class used by IQM_Vis.main.make_app to control all of the image widgets
class images:
    request_range_work = pyqtSignal(dict)
    view_correlation_instance = pyqtSignal(str)

    def __init__(self, **kwargs):
        self.update_images = True
        self.init_worker_thread()
        self.view_correlation_instance.connect(self.change_to_specific_trans)

    def init_worker_thread(self):
        ''' set up thread for smoother range plot calculation '''
        self.range_worker = IQM_Vis.UI.threads.get_range_results_worker()
        self.range_worker_thread = QThread()
        self.range_worker.progress.connect(self.update_progress)
        self.range_worker.current_image.connect(self.update_status_bar)
        self.range_worker.completed.connect(self.completed_range_results)
        self.range_worker.stopped.connect(self.stopped_range_worker)
        self.request_range_work.connect(self.range_worker.do_work)
        # move worker to the worker thread
        self.range_worker.moveToThread(self.range_worker_thread)
        # start the thread
        self.range_worker_thread.start()
        self.worker_working = False

    '''
    image updaters
    '''
    # def transform_image(self, image):
    #     for key in self.sliders['transforms']:
    #         image = self.sliders['transforms'][key]['function'](image, self.params_from_sliders['transforms'][key])
    #     return image

    def display_images(self):
        if self.update_images == True:
            for i, data_store in enumerate(self.data_stores):
                # reference image
                reference_image = data_store.get_reference_image()
                gui_utils.change_im(self.widget_row[i]['images']['original']['data'], reference_image, resize=self.image_display_size, rgb_brightness=self.rgb_brightness, display_brightness=self.display_brightness)
                # transform image
                trans_im = image_utils.get_transform_image(data_store, self.sliders['transforms'], self.params_from_sliders['transforms'])
                gui_utils.change_im(self.widget_row[i]['images']['transformed']['data'], trans_im, resize=self.image_display_size, rgb_brightness=self.rgb_brightness, display_brightness=self.display_brightness)
                # metrics
                metrics = data_store.get_metrics(trans_im, self.checked_metrics, **self.params_from_sliders['metric_params'])
                self.display_metrics(metrics, i)
                # metric images
                metric_images = data_store.get_metric_images(trans_im, self.checked_metric_images, **self.params_from_sliders['metric_params'])
                self.display_metric_images(metric_images, i)

                QApplication.processEvents()   # make figures become the correct size

    '''
    metric graph updaters
    '''
    def redo_plots(self, calc_range=False):
        if calc_range == True:  # add an or metrics_range val not been calculated
            if self.metrics_avg_graph or self.metric_range_graph:
                self.get_metrics_over_all_trans_with_init_values()

    '''
    change image in dataset
    '''
    def change_preview_images(self, ival):
        # have roll around scrolling
        if self.preview_num + ival < 0:
            self.preview_num = self.max_data_ind - self.num_images_scroll_show + 1
            if self.preview_num + ival < 0: # dataset is smaller than the number of images to show
                self.preview_num = 0
        elif self.preview_num + ival + self.num_images_scroll_show > self.max_data_ind + 1:
            self.preview_num = 0
        else:
            self.preview_num += ival
        self.set_preview_images(self.preview_num)
        
    def set_preview_images(self, preview_num):
        for data_store in self.data_stores:
            for i in range(min(self.num_images_scroll_show, self.max_data_ind+1)):
                im_preview_ind = preview_num + i
                if im_preview_ind <= self.max_data_ind:
                    # load image
                    im_preview_data = data_store.get_reference_image_by_index(im_preview_ind)
                    self.widget_im_num_hash[i] = im_preview_ind
                    # set image preview size
                    scale = self.num_images_scroll_show - 1
                    if isinstance(self.image_display_size, list) or isinstance(self.image_display_size, tuple):
                        preview_size = (self.image_display_size[0]//scale, self.image_display_size[1]//scale)
                    else:
                        preview_size = self.image_display_size//scale
                    # determine if we have the current display image
                    if self.data_num == im_preview_ind:
                        border = True
                    else:
                        border = False
                    gui_utils.change_im(self.widget_controls['images'][i], im_preview_data,
                                        resize=preview_size, rgb_brightness=self.rgb_brightness, display_brightness=self.display_brightness, border=border)
            self.widget_controls['label']['data_num'].setText(
                f'({preview_num+1}-{im_preview_ind+1}/{self.max_data_ind+1})')

    def change_data_click_im(self, widget_ind, *args): # args sent are position of mouse click on the image widget
        self.change_to_data_num(self.widget_im_num_hash[widget_ind])
        self.set_preview_images(self.preview_num) # refresh boarder on image

    def change_to_data_num(self, ind):  # change to an exact ind in the data list
        # check the num is within the data limits
        if ind < 0:
            return
        elif ind > self.max_data_ind:
            return
        else:
            self.data_num = ind
            self.change_data(0) # 0 means dont increment data_num

    def change_data(self, ival, _redo_plots=True):
        # reset any range/correlation data stored
        if hasattr(self, 'correlation_data'):
            self.correlation_data = {}
            for i, data_store in enumerate(self.data_stores):
                self.correlation_data[i] = {}
        self.data_num += ival
        # check the num is within the data limits
        if self.data_num < 0:
            self.data_num = 0
            return
        if self.data_num > self.max_data_ind:
            self.data_num = self.max_data_ind
            return
        self.range_worker.stop() # stop any calculations on the old image
        # if hasattr(self, 'wait_until_safe_to_change_image'):
        #     self.wait_until_safe_to_change_image()
        # if ival != 0:
        for data_store in self.data_stores:
            try:
                data_store[self.data_num]
            except:
                pass # some datasets will be shorter than others - this is fine though
            self.display_images()
            self.set_image_name_text()
        if _redo_plots == True:
            self.redo_plots(calc_range=True)
        # load human experiment if any
        self.human_experiment_scores = {}
        for i, data_store in enumerate(self.data_stores):
            # first try get from UI
            # load most recent correlation plots
            name = data_store.get_reference_image_name()
            if name in self.human_experiment_cache:
                self._load_experiment(self.human_experiment_cache[name], change_image=False)
            # otherwise from the datastore scores
            elif hasattr(data_store, 'human_scores'):
                self.human_experiment_scores[i] = data_store.human_scores

    def load_new_single_image(self):
        ''' change the image we are using '''
        # get the file opener for the user
        try:
            start_dir = os.path.expanduser("~")
            file, _ = QFileDialog.getOpenFileName(self,
                                               "Choose Image",
                                               start_dir,
                                               "All Files (*);; PNG Files (*.png)")
        except:
            return
        if file == ('', '') or file == '':
            return
        
        if os.path.isfile(file):
            image_format = imghdr.what(file)
            if image_format == None:
                self.update_status_bar(f'Not an image file: {file}', 10000)
            else:
                self.update_datastore_image_list([file])
        else:
            self.update_status_bar(f'Cannot find file: {file}', 10000)


    def load_new_images_folder(self):
        ''' change the image dataset we are using '''
        # get the file opener for the user
        try:
            start_dir = os.path.expanduser("~")
            dir = QFileDialog.getExistingDirectory(self, 
                                                   'Choose Image folder',
                                                   start_dir)
        except:
            return

        if dir == '':
            return

        image_list = glob.glob(os.path.join(dir, '*'))
        # remove and folders
        image_list = [f for f in image_list if os.path.isfile(f)]
        image_list.sort()

        self.update_datastore_image_list(image_list)

    def update_datastore_image_list(self, image_list, append_dataset=False): # onyl changes the first datastore
        # make sure that we have some images
        if image_list == []:
            self.update_status_bar('No images found', 10000)
            return
        # get all images if possible and we want to append images to the current dataset
        if append_dataset == True and hasattr(self.data_stores[0], 'get_image_dataset_list'):
            image_list = self.data_stores[0].get_image_dataset_list() + image_list
            data_num_to_change_to = len(self.data_stores[0])
        else:
            data_num_to_change_to = 0
        # change image dataset
        if hasattr(self.data_stores[0], 'load_image_list') and len(image_list) != 0:
            self.data_stores[0].load_image_list(image_list)
            self.max_data_ind = len(self.data_stores[0]) - 1
            self.data_num = data_num_to_change_to
            self.construct_UI()
        else:
            self.update_status_bar('Failed to update images', 10000)

    def load_human_experiment(self):
        ''' change the image dataset we are using '''
        # get the file opener for the user
        if os.path.exists(self.default_save_dir):
            start_dir = self.default_save_dir
        else:
            start_dir = os.path.expanduser("~")
        try:
            file, _ = QFileDialog.getOpenFileName(
                self,
                "Open Human Experiments File for Current Image",
                start_dir,
                "CSV Files (*.csv)",)
        except:
            return
        self._change_human_exp_2AFC(file)
        
    def _change_human_exp_2AFC(self, file, change_image=True):
        # load image
        if change_image == True:
            self._load_experiment_image(os.path.dirname(file))
        # load the csv human scores file and take mean of all experiments
        self.update_status_bar(f'Loading experiment file: {file}', 10000)
        if os.path.exists(file):
            df = pd.read_csv(file)
            for i, data_store in enumerate(self.data_stores):
                self.human_experiment_scores[i] = {'mean': df.mean().to_dict(),
                                                   'std': df.std().to_dict()}
                # cache this as the last used dataset for this image
                self.human_experiment_cache[data_store.get_reference_image_name()] = os.path.dirname(file)
            self.human_scores_file = file
        else:
            self.update_status_bar(f'No experiment file: {file}', 10000)
            return
        self._load_experiment_extras(os.path.dirname(file))
        self.update_status_bar(f'Loaded experiment file: {file}', 10000)

    def _change_human_exp_JND(self, file):
        # load the csv human scores file and take mean of all experiments
        self.update_status_bar(f'Loading experiment file: {file}', 10000)
        if os.path.exists(file):
            # now extract all the useful data from the csv
            df = pd.read_csv(file)
            # unique values for transform in df
            transforms = list(df['Transform'].unique())
            if 'None<>None' in transforms:
                transforms.remove('None<>None')
            trans_types = {}
            params = {}
            for single_trans in transforms:
                trans, trans_value = gui_utils.get_trans_dict_from_str(
                    single_trans)
                trans_types[trans] = 1 # just to get the keys
                params[trans_value] = 1 # just to get the keys

            # currently only one transformation type is supported
            if len(trans_types) > 1:
                self.update_status_bar(
                    'Warning: More than one transformation type in JND file', 10000)
                return
            
            transform_name = list(trans_types.keys())[0]
            self.current_JND_transform = transform_name
            params = list(params.keys())
            params.sort()

            # get the decisions for each parameter
            decisions = {}
            for param in params:
                name = save_utils.make_name_for_trans({'transform_name': transform_name, 
                                                       'transform_value': param})
                user_choices = df[df['Transform']
                                      == f'{name}']['UserDecision'].values
                # convert to plotable format
                values = []
                for choice in user_choices:
                    if choice == 'diff':
                        values.append(1)
                    elif choice == 'same':
                        values.append(0)
                    else:
                        values.append(-1)
                decisions[param] = values
            
            self.JND_trans = transform_name
            self.human_experiment_scores_JND = []
            for i, data_store in enumerate(self.data_stores):
                self.human_experiment_scores_JND.append(decisions)
            self.human_scores_JND_file = file
        else:
            self.update_status_bar(f'No experiment file: {file}', 10000)
            return
        self._load_experiment_extras_JND(os.path.dirname(file))
        self.update_status_bar(f'Loaded experiment file: {file}', 10000)

    def _load_experiment(self, dir, change_image=True):
        file = IQM_Vis.utils.save_utils.get_human_scores_file(dir)
        self._change_human_exp_2AFC(file, change_image=change_image)

    def _load_experiment_JND(self, dir):
        file = IQM_Vis.utils.save_utils.get_human_scores_file(dir)
        self._change_human_exp_JND(file)

    def load_experiment_from_dir(self):
        # get the file opener for the user
        try:
            start_dir = IQM_Vis.utils.save_utils.DEFAULT_SAVE_DIR
            dir = QFileDialog.getExistingDirectory(self, 
                                                   'Choose 2AFC Experiment Folder',
                                                   start_dir)
        except:
            return

        if dir == '':
            return   
        self._load_experiment(dir)

    def load_experiment_from_dir_JND(self):
        # get the file opener for the user
        try:
            start_dir = IQM_Vis.utils.save_utils.DEFAULT_SAVE_DIR
            dir = QFileDialog.getExistingDirectory(self, 
                                                   'Choose JND Experiment Folder',
                                                   start_dir)
        except:
            return

        if dir == '':
            return   
        self._load_experiment_JND(dir)

    def _load_experiment_image(self, dir):
        # load image if available already
        exp_image_name = IQM_Vis.utils.save_utils.get_image_name_from_human_scores(dir)
        exp_saved_image = IQM_Vis.utils.save_utils.get_original_unprocessed_image_file(dir)
        if hasattr(self.data_stores[0], 'image_names'):
            if exp_image_name in self.data_stores[0].image_names:
                ind = self.data_stores[0].image_names.index(exp_image_name)
                self.change_to_data_num(ind)
            # else load it from the experiment
            elif os.path.exists(exp_saved_image):
                self.update_datastore_image_list([exp_saved_image], append_dataset=True)
        # else load it from the experiment
        elif os.path.exists(exp_saved_image):
            self.update_datastore_image_list([exp_saved_image], append_dataset=True)
        else:
            self.update_status_bar(f'No {exp_image_name} in dataset or file {exp_saved_image}', 10000)
        
    def _load_experiment_extras(self, dir):
        # load the image processing if available
        processing_file = IQM_Vis.utils.save_utils.get_image_processing_file(dir)
        if os.path.exists(processing_file):
            processing = IQM_Vis.utils.save_utils.load_json_dict(
                processing_file)
            updated = False
            for name, options_var, change_func in zip(['pre', 'post'], 
                                                      [self.pre_processing_options, self.post_processing_options], 
                                                      [self.change_pre_processing, self.change_post_processing]):
                if processing[name] in options_var:
                    if processing[name] != self.widget_settings[f'image_{name}_processing']['widget'].currentText():
                        self.widget_settings[f'image_{name}_processing']['widget'].setCurrentText(
                            processing[name])
                        change_func()
                        updated = True
                else:
                    self.update_status_bar(
                        f'Could not load image setting: {processing[name]}')
            if updated == True:
                self.update_image_settings()
        else:
            self.update_status_bar(f'Warning: No settings file ({processing_file}), make sure image pre/post processing settings are correct')

        # clear cache and update correlation plot
        self.reset_correlation_data()
        self.display_metric_correlation_plot()
        self.tabs['graph'].setCurrentIndex(3)

    def _load_experiment_extras_JND(self, dir):
        # load the image processing if available
        processing_file = IQM_Vis.utils.save_utils.get_image_processing_file(dir)
        if os.path.exists(processing_file):
            processing = IQM_Vis.utils.save_utils.load_json_dict(
                processing_file)
            updated = False
            for name, options_var, change_func in zip(['pre', 'post'], 
                                                      [self.pre_processing_options, self.post_processing_options], 
                                                      [self.change_pre_processing, self.change_post_processing]):
                if processing[name] in options_var:
                    if processing[name] != self.widget_settings[f'image_{name}_processing']['widget'].currentText():
                        self.widget_settings[f'image_{name}_processing']['widget'].setCurrentText(
                            processing[name])
                        change_func()
                        updated = True
                else:
                    self.update_status_bar(
                        f'Could not load image setting: {processing[name]}')
            if updated == True:
                self.update_image_settings()
        else:
            self.update_status_bar(f'Warning: No settings file ({processing_file}), make sure image pre/post processing settings are correct')

        self.display_metric_JND_plot()
        self.tabs['graph'].setCurrentIndex(4)

    '''
    metric updaters
    '''
    def display_metrics(self, metrics, i):
        if self.metrics_info_format == 'graph':
            self.display_metrics_graph(metrics, i)
        else:
            self.display_metrics_text(metrics, i)

    def display_metrics_graph(self, metrics, i):
        bar_plt = plot_utils.bar_plotter(bar_names=[''],
                                        var_names=list(metrics.keys()),
                                        ax=self.widget_row[i]['metrics']['info']['data'],
                                        lim=self.plot_data_lim)
        bar_plt.plot('', list(metrics.values()))
        bar_plt.show()

    def display_metrics_text(self, metrics, i, disp_len=5):
        text = ''
        for key in metrics:
            metric = gui_utils.str_to_len(str(metrics[key]), disp_len, '0')
            text += key + ': ' + metric + '\n'
        self.widget_row[i]['metrics']['info']['data'].setText(text)

    '''
    get metric values when adjusting a single transformation value over its range
    '''
    def get_metrics_over_all_trans_with_init_values(self):
        # use the initiased/default values for all sliders
        # bundle up data needed to send to the worker
        data = {'trans': self.checked_transformations,
                'metric_params': self.params_from_sliders['metric_params'],
                'metrics_to_use': self.checked_metrics,
                'data_stores': self.data_stores,
                'num_steps': self.num_steps_range,
                }
        # start the worker working
        self.worker_working = True
        self.request_range_work.emit(data)

    def completed_range_results(self, results):
        ''' data results sent from signal from thread worker '''
        self.update_status_bar('Completed Range Results', 3000)
        self.stopped_range_worker()
        # first check if we have empty results (means we cannot plot) - have changed to accept empty
        # if 'metric_over_range_results' in results:
        #     if len(results['metric_over_range_results']) == 1:
        #         if results['metric_over_range_results'][0] == {}:
        #             # dont plot
        #             return 
        # else:
        #     # incorrect data
        #     return
        
        self.metric_over_range_results = results['metric_over_range_results']
        if 'max_val' in results:
            self.data_lims['range_data'] = results['max_val']
        if self.metrics_avg_graph:
            self.display_radar_plots()
        if self.metric_range_graph:
            self.display_metric_range_plot()
        self.display_metric_correlation_plot()

    '''
    metric range plot (line plots of range of all sliders)
    '''
    def display_metric_range_plot(self):
        if not hasattr(self, 'metric_over_range_results'):
            return
        all_trans = list(self.checked_transformations.keys())
        if all_trans == [] or self.checked_metrics == []:
            for i, data_store in enumerate(self.data_stores):
                axes = self.widget_row[i]['metrics']['range']['data']
                message_on_plot(axes, 'No Metrics/Transforms Selected')
            return
        trans_to_plot = all_trans[self.metric_range_graph_num]
        for i, data_store in enumerate(self.data_stores):
            if 'range' in self.widget_row[i]['metrics'].keys():
                axes = self.widget_row[i]['metrics']['range']['data']
                plot = plot_utils.get_transform_range_plots(self.metric_over_range_results[i], trans_to_plot, axes, self.plot_data_lim)
                plot.show()
    
    def plot_metric_range_mlp(self, i):
        # TODO: reduce copy of code from above function
        # make sure we have somethign to plot
        if not hasattr(self, 'metric_over_range_results'):
            return

        # plot our data on a new axis    
        fig = gui_utils.plt.figure()
        axes = fig.add_subplot(111)
        all_trans = list(self.checked_transformations.keys())
        trans_to_plot = all_trans[self.metric_range_graph_num]
        plot = plot_utils.get_transform_range_plots(
            self.metric_over_range_results[i], trans_to_plot, axes, self.plot_data_lim)
        plot.set_style()
        fig.canvas.manager.set_window_title(
            f"{self.data_stores[i].get_reference_image_name()}-{trans_to_plot}")

        self.set_save_dir_mpl(i)

        # show fig in new window
        gui_utils.matplotlib.pyplot.show()

    def change_metric_range_graph(self, add=1):
        max_graph_num = len(list(self.checked_transformations.keys()))
        self.metric_range_graph_num += add
        if self.metric_range_graph_num >= max_graph_num:
            self.metric_range_graph_num = max_graph_num - 1
        elif self.metric_range_graph_num < 0:
            self.metric_range_graph_num = 0
        self.display_metric_range_plot()

    '''
    metric averaging plots (radar plot)
    '''
    def display_radar_plots(self):
        for i, data_store in enumerate(self.data_stores):
            self.plot_radar_graph(self.metric_over_range_results[i], i)
            # uncomment below if you want to calc over the current trans values instead of init
            # results = plot_utils.compute_metrics_over_range(data_store,
            #                                                 self.checked_transformations,
            #                                                 self.params_from_sliders['transforms'],
            #                                                 self.params_from_sliders['metric_params'],
            #                                                 pbar=self.pbar)

    def plot_radar_graph(self, results, i):
        all_trans = list(self.checked_transformations.keys())
        if all_trans == [] or self.checked_metrics == []:
            for i, data_store in enumerate(self.data_stores):
                axes = self.widget_row[i]['metrics']['avg']['data']
                message_on_plot(axes, 'No Metrics/Transforms Selected')
            return
        if 'avg'  in self.widget_row[i]['metrics'].keys():
            # get current metrics used for this data_store
            metrics_names = []
            for metric in self.data_stores[i].metrics:
                if metric in self.checked_metrics:
                    metrics_names.append(metric)

            if len(metrics_names) == 0:
                return

            transformation_names = list(self.sliders['transforms'].keys())
            axes = self.widget_row[i]['metrics']['avg']['data']
            radar_plotter = plot_utils.get_radar_plots_avg_plots(results, metrics_names, transformation_names, axes, self.plot_data_lim)
            radar_plotter.show()
    
    def plot_radar_mlp(self, i):
        # TODO: reduce copy of code from above function
        # make sure we have somethign to plot
        if not hasattr(self, 'metric_over_range_results'):
            return

        # plot our data on a new axis
        fig = gui_utils.plt.figure()
        axes = fig.add_subplot(projection='polar')
        all_trans = list(self.checked_transformations.keys())
        transformation_names = list(self.sliders['transforms'].keys())
        if 'avg' in self.widget_row[i]['metrics'].keys():
            # get current metrics used for this data_store
            metrics_names = []
            for metric in self.data_stores[i].metrics:
                if metric in self.checked_metrics:
                    metrics_names.append(metric)

            if len(metrics_names) == 0:
                return
            
        plot = plot_utils.get_radar_plots_avg_plots(
            self.metric_over_range_results[i], metrics_names, transformation_names, axes, self.plot_data_lim)
        plot.set_style()

        fig.canvas.manager.set_window_title(
            f"{self.data_stores[i].get_reference_image_name()}-radar")
        
        self.set_save_dir_mpl(i)

        # show fig in new window
        gui_utils.matplotlib.pyplot.show()

    def get_export_dir(self, i):
        image_name = f"{self.data_stores[i].get_reference_image_name()}"
        save_path = os.path.join(self.default_save_dir, 'exports', image_name)
        return save_path

    def set_save_dir_mpl(self, i=0):
        # set the default save path
        save_path = self.get_export_dir(i)
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        gui_utils.matplotlib.rcParams['savefig.directory'] = save_path

    '''
    metric correlation plots
    '''
    def display_metric_JND_plot(self):
        for i, data_store in enumerate(self.data_stores):
            axes = self.widget_row[i]['metrics']['JND']['data']
            all = {} # to calucalte std of each 
            for val, decisions in self.human_experiment_scores_JND[i].items():
                for decision in decisions:
                    if val in all.keys():
                        all[val].append(decision)
                    else:
                        all[val] = [decision]

            # calculate mean and std decision of each val
            x = []
            y = []
            e = []
            labels = []
            for val, decisions in all.items():
                x.append(val)
                y.append(np.mean(decisions))
                e.append(np.std(decisions))
                labels.append(f'{self.JND_trans}={val}')

            sp = plot_utils.scatter_plotter(axes,
                                            x_label=f'{self.current_JND_transform} Value',
                                            y_label='User Decision')
            sp.plot(x, y, annotations=labels, error=e)

            # # make interactive hover for points
            # annot = sp.ax.axes.annotate("", xy=(0, 0), xytext=(0, 0), textcoords="offset points",
            #                             bbox=dict(boxstyle="round", fc="w"),
            #                             arrowprops=dict(arrowstyle="->")
            #                             )
            # annot.set_visible(False)
            # sp.ax.figure.canvas.mpl_connect(
            #     "motion_notify_event", partial(plot_utils.hover_scatter, sp, annot))
            # sp.ax.figure.canvas.mpl_connect(
            #     "pick_event", partial(plot_utils.click_scatter, sp, plot_utils.change_trans_value_signal))

            
            # sp = plot_utils.scatter_plotter(axes,
            #                                 x_label=f'{self.current_JND_transform} Value',
            #                                 y_label='User Decision')
            # sp.plot(x, y)
            sp.show()

            # ax.scatter(x, y)
            # ax.set_xlabel(f'{self.current_JND_transform} Value')
            # ax.set_ylabel('User Decision')
            # # ax.set_title(f'{self.current_JND_transform} vs User Decision')
            # plot.show()

    def display_metric_correlation_plot(self):
        if self.checked_metrics == []:
            for i, data_store in enumerate(self.data_stores):
                axes = self.widget_row[i]['metrics']['correlation']['data']
                message_on_plot(axes, 'No Metrics Loaded')
            return
        metric = self.checked_metrics[self.metric_correlation_graph_num]
        # calculate the metric values at the human score test values
        for i, data_store in enumerate(self.data_stores):
            if i in self.human_experiment_scores.keys():
                # calculate metric at HIQM values (if not already cached)
                if metric not in self.correlation_data[i].keys():
                    IQM_scores = plot_utils.compute_metric_for_human_correlation(data_store,
                                            self.checked_transformations,
                                            self.params_from_sliders['metric_params'],
                                            trans_str_values = self.human_experiment_scores[i]['mean'].keys(),
                                            metric=metric)
                    # cache it
                    self.correlation_data[i][metric] = IQM_scores
                plot = plot_utils.get_correlation_plot(self.human_experiment_scores[i],
                                                self.correlation_data[i],
                                                self.widget_row[i]['metrics']['correlation']['data'],
                                                metric,
                                                self.view_correlation_instance)
                if hasattr(self, 'human_scores_file'):
                    title = plot.ax.axes.get_title()
                    title = f"{title}\n{self.human_scores_file}"
                    plot.ax.axes.set_title(title, fontsize=8)
                plot.show()
            else:
                axes = self.widget_row[i]['metrics']['correlation']['data']
                message_on_plot(axes, 'Use Load Experiment Button below\nor File->Load Human Scores for a .csv')
                

    def change_metric_correlations_graph(self, add=1):
        max_graph_num = len(self.checked_metrics)
        self.metric_correlation_graph_num += add
        if self.metric_correlation_graph_num >= max_graph_num:
            self.metric_correlation_graph_num = max_graph_num - 1
        elif self.metric_correlation_graph_num < 0:
            self.metric_correlation_graph_num = 0
        self.display_metric_correlation_plot()

    def change_to_specific_trans(self, trans_str):
        trans, trans_value = gui_utils.get_trans_dict_from_str(trans_str)
        # load transform if not already on the UI
        trans_found = False
        for _, slider_group in self.sliders.items():
            for key, item_sliders in slider_group.items():
                if key == trans:
                    trans_found = True
        if trans_found == False: # need to load it
            all_trans_iqm_vis = IQM_Vis.transforms.get_all_transforms()
            if trans in self.menu_options['transforms']:
                self.menu_options['transforms'][trans].setChecked(True)
                self.construct_UI()
            elif trans in all_trans_iqm_vis:
                self.transformations[trans] = all_trans_iqm_vis[trans]
                self._remake_menu()
                self.menu_options['transforms'][trans].setChecked(True)
                self.construct_UI()
            else:
                self.update_status_bar(f"Transform '{trans}' not found - please load it!")

        # show the specific tranform image
        for _, slider_group in self.sliders.items():
            for key, item_sliders in slider_group.items():
                if key == trans:
                    closest_slider_val = np.argmin(np.abs(self.sliders['transforms'][key]['values']-trans_value))
                    self.widget_controls['slider'][key]['data'].setValue(closest_slider_val)
                else:
                    self.widget_controls['slider'][key]['data'].setValue(item_sliders['init_ind'])
        self.display_images()
        self.redo_plots(calc_range=False)


    '''
    metric image updaters
    '''
    def display_metric_images(self, metric_images, i):
        for key in metric_images:
            widget = self.widget_row[i]['metric_images'][key]['data']
            gui_utils.change_im(widget, metric_images[key], resize=self.image_display_size)

    '''
    thread manegment
    '''
    def stopped_range_worker(self, signal=False):
        self.worker_working = False

def message_on_plot(widget_axes, message):
    try:
        widget_axes.axes.clear()
    except AttributeError:
        pass
    widget_axes.axes.text(
        0.5, 0.5, message, horizontalalignment='center', verticalalignment='center')
    widget_axes.draw()


# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License
'''
Helper functions for metric computation
'''
import numpy as np
import torch

def _numpy_to_torch_image(image):
    if len(image.shape) == 2:
        image = np.expand_dims(image, axis=2)
    image = image.transpose((2, 0, 1))
    image = np.expand_dims(image, axis=0)   # make into batch one 1
    image = torch.from_numpy(image)
    return image


def _check_shapes(im_ref, im_comp):
    '''
    make sure both images have the same dimensions
    '''
    if im_ref.shape != im_comp.shape:
        raise ValueError(
            f'Refence and transformed images need to have the same shape not: {im_ref.shape} and {im_comp.shape}')



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
import glob
import numpy as np
import torch
import IQM_Vis

from PIL import Image
from PIL.TiffTags import TAGS



'''Textures calibrated image loader'''
def load_and_calibrate_image(file, max_luminance=200, size=512):
    # Calculate max luminance value in the whole dataset
    img = Image.open(file)
    meta_dict = {TAGS[key] : img.tag[key] for key in img.tag_v2}
    img = np.array(img).astype(np.float64)
    lms = correct(img, meta_dict)

    # Get each point in rgb space to visualise the colour in RGB
    # XYZ -> LMS
    Mxyzlms = np.array([ # Bradford LMS transform
        [0.8951000, 0.2664000, -0.1614000],
        [-0.7502000, 1.7135000, 0.0367000],
        [0.0389000, -0.0685000, 1.0296000]])
    # LMS -> XYZ
    Mlmsxyz = np.linalg.inv(Mxyzlms)

    # RGB -> XYZ
    Mng2xyz = np.array([[69.1661, 52.4902, 46.6052],
                        [39.0454, 115.8404, 16.3118],
                        [3.3467, 12.6700, 170.1090]])
    #XYZ -> RGB
    Mxyz2ng = np.linalg.inv(Mng2xyz)

    # Convert LMS -> XYZ and scale values for max luminance default: 200 cd/m^2
    xyz_corrected = (lms @ Mlmsxyz.T)/np.max(lms[:, 0]) * max_luminance

    # Convert XYZ -> RGB
    rgb_corrected = np.clip(xyz_corrected @ Mxyz2ng.T, 0.0, 1.0)
    rgb_small = IQM_Vis.utils.resize_to_longest_side(rgb_corrected, side=size)
    return rgb_small

''' util for calibrating images'''
# camera correction and coversation into LMS colourspace
def correct(img, meta_dict, greycale=True):
    if 'merry' in meta_dict['ImageDescription'][0]:
        T = np.array([
            [0.428443253, 0.495562896, 0.075993851],
            [0.243026144, 0.614128681, 0.142845175],
            [0.155766424, 0.132343175, 0.711890401]])
        a_R = 7.320565961
        a_G = 12.0579051
        a_B = 10.6112984
        b = 1.008634316
    elif 'pippi' in meta_dict['ImageDescription'][0]:
        T = np.array([
            [0.431088433, 0.494438389, 0.074473178],
            [0.245488691, 0.614786761, 0.139724548],
            [0.166472303, 0.124487321, 0.709040376]])
        a_R = 5.562441185
        a_G = 8.876002262
        a_B = 7.233814813
        b = 1.009696031
    II = np.zeros_like(img)
    II[:, :, 0] = a_R*(b**img[:, :, 0]-1)
    II[:, :, 1] = a_G*(b**img[:, :, 1]-1)
    II[:, :, 2] = a_B*(b**img[:, :, 2]-1)
    exposure_time = meta_dict['ImageDescription'][0].split('-')[-1]
    try:
        exposure_time = np.float64(exposure_time)
    except:
        return None
    II = II / np.float64(exposure_time)
    II = np.reshape(np.reshape(II, (np.prod(II.shape[0:2]), 3), order='F')@T.T, (II.shape), order='F')
    return II

def run():
    # metrics functions must return a single value
    metric = {'DISTS': IQM_Vis.metrics.DISTS(),
              'LPIPS': IQM_Vis.metrics.LPIPS(),
              '1-SSIM': IQM_Vis.metrics.SSIM(),
              'MAE': IQM_Vis.metrics.MAE(),
            #   '1-MS_SSIM': IQM_Vis.metrics.MS_SSIM(),
            #   'NLPD': IQM_Vis.metrics.NLPD(),
              }

    # metrics images return a numpy image - dont include any for this example
    metric_images = {}

    # make dataset list of images
    files = sorted(glob.glob(os.path.join(os.path.expanduser('~'), 'datasets', 'Textures', '*')))
    if len(files) == 0:
        raise ValueError('No files found in the dataset directory - please download the texture dataset')

    data = IQM_Vis.dataset_holder(files,
                                  metric,
                                  metric_images,
                                  image_loader=load_and_calibrate_image,
                                  image_pre_processing=lambda x: x,  # no resize at start up 
                                  image_post_processing=lambda x: IQM_Vis.utils.image_utils.crop_centre(
                                      x, scale_factor=2, keep_size=False)
                                  )

    # define the transformations
    transformations = {
        'rotation':{'min':-12, 'max':12, 'function':IQM_Vis.transforms.rotation},    # normal input
        'x_shift': {'min':-0.2, 'max':0.2, 'function':IQM_Vis.transforms.x_shift, 'init_value': 0.0},
        # 'y_shift': {'min':-0.1, 'max':0.1, 'function':IQM_Vis.transforms.y_shift, 'init_value': 0.0},
        'zoom':    {'min': 0.8, 'max':1.2, 'function':IQM_Vis.transforms.zoom_image, 'init_value': 1.0},  # requires non standard slider params
        # 'brightness':{'min':-1.0, 'max':1.0, 'function':IQM_Vis.transforms.brightness},   # normal but with float
        # 'contrast': {'min': 0.5, 'max': 2.5, 'init_value': 1.0, 'function': IQM_Vis.transforms.contrast},
        # 'hue': {'min': -0.5, 'max': 0.5, 'function': IQM_Vis.transforms.hue},
        # 'saturation': {'min': -0.5, 'max': 0.5, 'function': IQM_Vis.transforms.saturation},
        # 'jpg compr':{'init_value':101, 'min':1, 'max':101, 'function':IQM_Vis.transforms.jpeg_compression},
        # 'blur':{'min':1, 'max':41, 'normalise':'odd', 'function':IQM_Vis.transforms.blur},  # only odd ints
        # 'threshold':{'min':-40, 'max':40, 'function':IQM_Vis.transforms.binary_threshold},
               }

    # use the API to create the UI
    IQM_Vis.make_UI(data, transformations)


if __name__ == '__main__':
    run()
